# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The server-side functions that perform job endpoint operations."""
from __future__ import annotations

from typing import Any, Final, cast

import structlog
from flask_login import current_user
from injector import inject
from sqlalchemy import func, select
from sqlalchemy.orm import aliased
from structlog.stdlib import BoundLogger

from dioptra.restapi.db import db, models
from dioptra.restapi.errors import (
    BackendDatabaseError,
    EntityDoesNotExistError,
    EntityNotRegisteredError,
    JobInvalidParameterNameError,
    JobInvalidStatusTransitionError,
    JobMlflowRunAlreadySetError,
    SortParameterValidationError,
)
from dioptra.restapi.v1 import utils
from dioptra.restapi.v1.entrypoints.service import (
    RESOURCE_TYPE as ENTRYPOINT_RESOURCE_TYPE,
)
from dioptra.restapi.v1.entrypoints.service import EntrypointIdService
from dioptra.restapi.v1.experiments.service import (
    RESOURCE_TYPE as EXPERIMENT_RESOURCE_TYPE,
)
from dioptra.restapi.v1.experiments.service import ExperimentIdService
from dioptra.restapi.v1.groups.service import GroupIdService
from dioptra.restapi.v1.queues.service import RESOURCE_TYPE as QUEUE_RESOURCE_TYPE
from dioptra.restapi.v1.queues.service import QueueIdService
from dioptra.restapi.v1.shared.rq_service import RQServiceV1
from dioptra.restapi.v1.shared.search_parser import construct_sql_query_filters

LOGGER: BoundLogger = structlog.stdlib.get_logger()

RESOURCE_TYPE: Final[str] = "job"
SEARCHABLE_FIELDS: Final[dict[str, Any]] = {
    "description": lambda x: models.Job.description.like(x),
    "status": lambda x: models.Job.status.like(x),
    "timeout": lambda x: models.Job.timeout.like(x),
    "tag": lambda x: models.Job.tags.any(models.Tag.name.like(x, escape="/")),
}
SORTABLE_FIELDS: Final[dict[str, Any]] = {
    "description": models.Job.description,
    "createdOn": models.Job.created_on,
    "lastModifiedOn": models.Resource.last_modified_on,
    "status": models.Job.status,
}
JOB_STATUS_TRANSITIONS: Final[dict[str, Any]] = {
    "queued": {"started", "deferred"},
    "started": {"finished", "failed"},
    "deferred": {"started"},
    "failed": {},
    "finished": {},
}


class JobService(object):
    """The service methods for registering and managing jobs by their unique id."""

    @inject
    def __init__(
        self,
        experiment_id_service: ExperimentIdService,
        queue_id_service: QueueIdService,
        entrypoint_id_service: EntrypointIdService,
        group_id_service: GroupIdService,
        rq_service: RQServiceV1,
    ) -> None:
        """Initialize the job service.

        All arguments are provided via dependency injection.

        Args:
            experiment_id_service: An ExperimentIdService object.
            queue_id_service: A QueueIdService object.
            entrypoint_id_service: An EntrypointIdService object.
            group_id_service: A GroupIdService object.
            rq_service: An RQServiceV1 object.
        """
        self._experiment_id_service = experiment_id_service
        self._queue_id_service = queue_id_service
        self._entrypoint_id_service = entrypoint_id_service
        self._group_id_service = group_id_service
        self._rq_service = rq_service

    def create(
        self,
        experiment_id: int,
        queue_id: int,
        entrypoint_id: int,
        values: dict[str, str],
        description: str,
        timeout: str,
        **kwargs,
    ) -> utils.JobDict:
        """Create a new job.

        Args:
            experiment_id: The unique id for the experiment this job is under.
            queue_id: The unique id for the queue this job will execute on.
            entrypoint_id: The unique id for the entrypoint defining the job.
            values: The parameter values passed to the entrypoint to configure the job.
            description: The description of the job.
            timeout: The length of time the job will run before timing out.
            group_id: The group that will own the job.

        Returns:
            The newly created job object.

        Raises:
            EntityExistsError: If a job with the given name already exists.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        # Set the default status
        status = "queued"

        # Validate the provided experiment_id and fetch the ORM object
        experiment_dict = cast(
            utils.ExperimentDict,
            self._experiment_id_service.get(
                experiment_id, error_if_not_found=True, log=log
            ),
        )
        experiment = experiment_dict["experiment"]

        # Validate that the provided entrypoint_id is registered to the experiment
        parent_experiment = aliased(models.Experiment)
        experiment_entry_point_ids_stmt = (
            select(models.EntryPoint.resource_id)
            .join(
                models.resource_dependencies_table,
                models.EntryPoint.resource_id
                == models.resource_dependencies_table.c.child_resource_id,
            )
            .join(
                parent_experiment,
                parent_experiment.resource_id
                == models.resource_dependencies_table.c.parent_resource_id,
            )
            .where(parent_experiment.resource_id == experiment_id)
        )
        experiment_entry_point_ids = list(
            db.session.scalars(experiment_entry_point_ids_stmt).all()
        )

        if entrypoint_id not in set(experiment_entry_point_ids):
            raise EntityNotRegisteredError(
                EXPERIMENT_RESOURCE_TYPE,
                experiment_id,
                ENTRYPOINT_RESOURCE_TYPE,
                entrypoint_id,
            )

        # Validate that the provided queue_id is registered to the entrypoint
        parent_entry_point = aliased(models.EntryPoint)
        entry_point_queue_ids_stmt = (
            select(models.Queue.resource_id)
            .join(
                models.resource_dependencies_table,
                models.Queue.resource_id
                == models.resource_dependencies_table.c.child_resource_id,
            )
            .join(
                parent_entry_point,
                parent_entry_point.resource_id
                == models.resource_dependencies_table.c.parent_resource_id,
            )
            .where(parent_entry_point.resource_id.in_(experiment_entry_point_ids))
        )
        entry_point_queue_ids = list(
            db.session.scalars(entry_point_queue_ids_stmt).all()
        )

        if queue_id not in set(entry_point_queue_ids):
            raise EntityNotRegisteredError(
                ENTRYPOINT_RESOURCE_TYPE, entrypoint_id, QUEUE_RESOURCE_TYPE, queue_id
            )

        # Fetch the validated queue
        queue_dict = cast(
            utils.QueueDict,
            self._queue_id_service.get(queue_id, error_if_not_found=True, log=log),
        )
        queue = queue_dict["queue"]

        # Fetch the validated entrypoint
        entrypoint_dict = cast(
            utils.EntrypointDict,
            self._entrypoint_id_service.get(
                entrypoint_id, error_if_not_found=True, log=log
            ),
        )
        entrypoint = entrypoint_dict["entry_point"]

        # Validate the keys in values against the registered entrypoint parameter names
        invalid_job_params = set(values.keys()) - set(
            param.name for param in entrypoint.parameters
        )
        if len(invalid_job_params) > 0:
            log.debug("Invalid parameter names", parameters=list(invalid_job_params))
            raise JobInvalidParameterNameError

        # Create the new Job resource and record the assigned entrypoint parameter
        # values
        job_resource = models.Resource(
            resource_type=RESOURCE_TYPE, owner=experiment.resource.owner
        )
        entrypoint_parameter_values = [
            models.EntryPointParameterValue(
                value=values.get(
                    entrypoint_parameter.name, entrypoint_parameter.default_value
                ),
                job_resource=job_resource,
                parameter=entrypoint_parameter,
            )
            for entrypoint_parameter in entrypoint.parameters
        ]

        new_job = models.Job(
            timeout=timeout,
            status=status,
            description=description,
            resource=job_resource,
            creator=current_user,
        )
        new_job.entry_point_job = models.EntryPointJob(
            job_resource=job_resource,
            entry_point=entrypoint,
            entry_point_parameter_values=entrypoint_parameter_values,
        )
        new_job.experiment_job = models.ExperimentJob(
            job_resource=job_resource,
            experiment=experiment,
        )
        new_job.queue_job = models.QueueJob(
            job_resource=job_resource,
            queue=queue,
        )
        db.session.add(new_job)
        db.session.commit()
        self._rq_service.submit(
            job_id=new_job.resource_id,
            experiment_id=experiment_id,
            queue=queue.name,
            timeout=timeout,
        )
        log.debug(
            "Job registration successful",
            job_id=new_job.resource_id,
        )
        return utils.JobDict(
            job=new_job,
            artifacts=[],
            has_draft=False,
        )

    def get(
        self,
        group_id: int | None,
        search_string: str,
        page_index: int,
        page_length: int,
        sort_by_string: str,
        descending: bool,
        **kwargs,
    ) -> tuple[list[utils.JobDict], int]:
        """Fetch a list of jobs, optionally filtering by search string and paging
        parameters.

        Args:
            group_id: A group ID used to filter results.
            search_string: A search string used to filter results.
            page_index: The index of the first group to be returned.
            page_length: The maximum number of jobs to be returned.
            sort_by_string: The name of the column to sort.
            descending: Boolean indicating whether to sort by descending or not.

        Returns:
            A tuple containing a list of jobs and the total number of jobs matching
            the query.

        Raises:
            BackendDatabaseError: If the database query returns a None when counting
                the number of jobs.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get full list of jobs")

        filters = list()

        if group_id is not None:
            filters.append(models.Resource.group_id == group_id)

        if search_string:
            filters.append(
                construct_sql_query_filters(search_string, SEARCHABLE_FIELDS)
            )

        stmt = (
            select(func.count(models.Job.resource_id))
            .join(models.Resource)
            .where(
                *filters,
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id == models.Job.resource_snapshot_id,
            )
        )
        total_num_jobs = db.session.scalars(stmt).first()

        if total_num_jobs is None:
            log.error(
                "The database query returned a None when counting the number of "
                "groups when it should return a number.",
                sql=str(stmt),
            )
            raise BackendDatabaseError

        if total_num_jobs == 0:
            return [], total_num_jobs

        jobs_stmt = (
            select(models.Job)
            .join(models.Resource)
            .where(
                *filters,
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id == models.Job.resource_snapshot_id,
            )
            .offset(page_index)
            .limit(page_length)
        )

        if sort_by_string and sort_by_string in SORTABLE_FIELDS:
            sort_column = SORTABLE_FIELDS[sort_by_string]
            if descending:
                sort_column = sort_column.desc()
            else:
                sort_column = sort_column.asc()
            jobs_stmt = jobs_stmt.order_by(sort_column)
        elif sort_by_string and sort_by_string not in SORTABLE_FIELDS:
            raise SortParameterValidationError(RESOURCE_TYPE, sort_by_string)

        jobs = list(db.session.scalars(jobs_stmt).all())

        job_dicts: dict[int, utils.JobDict] = {
            job.resource_id: utils.JobDict(
                job=job,
                artifacts=[],
                has_draft=False,
            )
            for job in jobs
        }

        job_ids = [job.resource_id for job in jobs]
        artifacts_stmt = (
            select(models.Artifact)
            .join(models.Resource)
            .where(
                models.Artifact.job_id.in_(job_ids),
                models.Resource.is_deleted == False,  # noqa: E712
            )
        )
        for artifact in db.session.scalars(artifacts_stmt).all():
            job_dicts[artifact.job_id]["artifacts"].append(artifact)

        return list(job_dicts.values()), total_num_jobs


class JobIdService(object):
    """The service methods for registering and managing jobs by their unique id."""

    def get(
        self,
        job_id: int,
        error_if_not_found: bool = False,
        **kwargs,
    ) -> utils.JobDict | None:
        """Fetch a job by its unique id.

        Args:
            job_id: The unique id of the job.
            error_if_not_found: If True, raise an error if the job is not found.
                Defaults to False.

        Returns:
            The job object if found, otherwise None.

        Raises:
            EntityDoesNotExistError: If the job is not found and `error_if_not_found`
                is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get job by id", job_id=job_id)

        stmt = (
            select(models.Job)
            .join(models.Resource)
            .where(
                models.Job.resource_id == job_id,
                models.Job.resource_snapshot_id == models.Resource.latest_snapshot_id,
                models.Resource.is_deleted == False,  # noqa: E712
            )
        )
        job = db.session.scalars(stmt).first()

        if job is None:
            if error_if_not_found:
                raise EntityDoesNotExistError(RESOURCE_TYPE, job_id=job_id)

            return None

        artifacts_stmt = (
            select(models.Artifact)
            .join(models.Resource)
            .where(
                models.Artifact.job_id == job_id,
                models.Resource.is_deleted == False,  # noqa: E712
            )
        )
        artifacts = list(db.session.scalars(artifacts_stmt).all())

        return utils.JobDict(
            job=job,
            artifacts=artifacts,
            has_draft=False,
        )

    def delete(self, job_id: int, **kwargs) -> dict[str, Any]:
        """Delete a job.

        Args:
            job_id: The unique id of the job.

        Returns:
            A dictionary reporting the status of the request.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        stmt = select(models.Resource).filter_by(
            resource_id=job_id, resource_type=RESOURCE_TYPE, is_deleted=False
        )
        job_resource = db.session.scalars(stmt).first()

        if job_resource is None:
            raise EntityDoesNotExistError(RESOURCE_TYPE, job_id=job_id)

        deleted_resource_lock = models.ResourceLock(
            resource_lock_type="delete",
            resource=job_resource,
        )
        db.session.add(deleted_resource_lock)
        db.session.commit()
        log.debug("Job deleted", job_id=job_id)

        return {"status": "Success", "id": [job_id]}


class JobIdStatusService(object):
    """The service methods for retrieving the status of a job by unique id."""

    @inject
    def __init__(
        self,
        job_id_service: JobIdService,
    ) -> None:
        """Initialize the job status service.

        All arguments are provided via dependency injection.

        Args:
            job_id_service: A JobIdService object.
        """
        self._job_id_service = job_id_service

    def get(
        self,
        job_id: int,
        **kwargs,
    ) -> dict[str, Any]:
        """Fetch a job's status by its unique id.

        Args:
            job_id: The unique id of the job.

        Returns:
            The status message job object if found, otherwise an error message.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get job status by id", job_id=job_id)

        stmt = (
            select(models.Job)
            .join(models.Resource)
            .where(
                models.Job.resource_id == job_id,
                models.Job.resource_snapshot_id == models.Resource.latest_snapshot_id,
                models.Resource.is_deleted == False,  # noqa: E712
            )
        )
        job = db.session.scalars(stmt).first()

        if job is None:
            raise EntityDoesNotExistError(RESOURCE_TYPE, job_id=job_id)

        return {"status": job.status, "id": job.resource_id}


class ExperimentJobService(object):
    """The service methods for submitting and retrieving jobs within an experiment
    namespace."""

    @inject
    def __init__(
        self, experiment_id_service: ExperimentIdService, job_service: JobService
    ) -> None:
        """Initialize the ExperimentIdJob service.

        All arguments are provided via dependency injection.

        Args:
            experiment_id_service: An ExperimentIdService object.
            job_service: A JobService object.
        """
        self._experiment_id_service = experiment_id_service
        self._job_service = job_service

    def create(
        self,
        experiment_id: int,
        queue_id: int,
        entrypoint_id: int,
        values: dict[str, str],
        description: str,
        timeout: str,
        **kwargs,
    ) -> utils.JobDict:
        """Create a new job within an experiment.

        Args:
            experiment_id: The unique id for the experiment this job is under.
            queue_id: The unique id for the queue this job will execute on.
            entrypoint_id: The unique id for the entrypoint defining the job.
            values: The parameter values passed to the entrypoint to configure the job.
            description: The description of the job.
            timeout: The length of time the job will run before timing out.
            group_id: The group that will own the job.

        Returns:
            The newly created job object.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        return self._job_service.create(
            experiment_id=experiment_id,
            queue_id=queue_id,
            entrypoint_id=entrypoint_id,
            values=values,
            description=description,
            timeout=timeout,
            log=log,
        )

    def get(
        self,
        experiment_id: int,
        search_string: str,
        page_index: int,
        page_length: int,
        sort_by_string: str,
        descending: bool,
        **kwargs,
    ) -> tuple[list[utils.JobDict], int]:
        """Fetch a list of jobs for a given experiment, optionally filtering by search
        string and paging parameters.

        Args:
            experiment_id: The unique id of the experiment.
            search_string: A search string used to filter results.
            page_index: The index of the first page to be returned.
            page_length: The maximum number of experiments to be returned.
            sort_by_string: The name of the column to sort.
            descending: Boolean indicating whether to sort by descending or not.

        Returns:
            A tuple containing a list of jobs and the total number of jobs matching the
            query.

        Raises:
            EntityDoesNotExistError: If the experiment is not found.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get full list of jobs for experiment", experiment_id=experiment_id)

        filters = []

        if search_string:
            filters.append(
                construct_sql_query_filters(search_string, SEARCHABLE_FIELDS)
            )

        cte_job_ids = (
            select(models.ExperimentJob.job_resource_id)
            .join(models.Resource)
            .where(
                models.ExperimentJob.experiment_id == experiment_id,
                models.Resource.is_deleted == False,  # noqa: E712
            )
            .cte()
        )
        stmt = (
            select(func.count(models.Job.resource_id))
            .join(models.Resource)
            .where(
                *filters,
                models.Job.resource_id.in_(select(cte_job_ids)),
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id == models.Job.resource_snapshot_id,
            )
        )
        total_num_jobs = db.session.scalars(stmt).first()

        if total_num_jobs is None:
            log.error(
                "The database query returned a None when counting the number of "
                "experiment jobs when it should return a number.",
                sql=str(stmt),
            )
            raise BackendDatabaseError

        if total_num_jobs == 0:
            return [], total_num_jobs

        jobs_stmt = (
            select(models.Job)
            .join(models.Resource)
            .where(
                *filters,
                models.Job.resource_id.in_(select(cte_job_ids)),
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id == models.Job.resource_snapshot_id,
            )
            .offset(page_index)
            .limit(page_length)
        )

        if sort_by_string and sort_by_string in SORTABLE_FIELDS:
            sort_column = SORTABLE_FIELDS[sort_by_string]
            if descending:
                sort_column = sort_column.desc()
            else:
                sort_column = sort_column.asc()
            jobs_stmt = jobs_stmt.order_by(sort_column)
        elif sort_by_string and sort_by_string not in SORTABLE_FIELDS:
            raise SortParameterValidationError(RESOURCE_TYPE, sort_by_string)

        jobs = list(db.session.scalars(jobs_stmt).all())

        job_dicts: dict[int, utils.JobDict] = {
            job.resource_id: utils.JobDict(
                job=job,
                artifacts=[],
                has_draft=False,
            )
            for job in jobs
        }

        return list(job_dicts.values()), total_num_jobs


class ExperimentJobIdService(object):
    """The service methods for getting or deleting a specific job within an experiment
    namespace."""

    @inject
    def __init__(
        self, experiment_id_service: ExperimentIdService, job_id_service: JobIdService
    ) -> None:
        """Initialize the ExperimentIdJob service.

        All arguments are provided via dependency injection.

        Args:
            experiment_id_service: An ExperimentIdService object.
            job_service: A JobService object.
        """
        self._experiment_id_service = experiment_id_service
        self._job_id_service = job_id_service

    def get(self, experiment_id: int, job_id: int, **kwargs) -> utils.JobDict:
        """Fetch a job by its unique id under a given experiment namespace.

        Args:
            experiment_id: The unique id of the experiment.
            job_id: The unique id of the job.

        Returns:
            The job object if found, otherwise None.

        Raises:
            EntityDoesNotExistError: If the job associated with the experiment
                is not found.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        experiment_job_stmt = select(models.ExperimentJob).where(
            models.ExperimentJob.experiment_id == experiment_id,
            models.ExperimentJob.job_resource_id == job_id,
        )
        experiment_job = db.session.scalar(experiment_job_stmt)

        if experiment_job is None:
            raise EntityDoesNotExistError(
                RESOURCE_TYPE, job_id=job_id, experiment_id=experiment_id
            )

        return cast(
            utils.JobDict,
            self._job_id_service.get(
                job_id=job_id,
                error_if_not_found=True,
                log=log,
            ),
        )

    def delete(self, experiment_id: int, job_id: int, **kwargs) -> dict[str, Any]:
        """Delete a job.

        Args:
            experiment_id: The unique id of the experiment.
            job_id: The unique id of the job.

        Returns:
            A dictionary reporting the status of the request.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        experiment_job_stmt = select(models.ExperimentJob).where(
            models.ExperimentJob.experiment_id == experiment_id,
            models.ExperimentJob.job_resource_id == job_id,
        )
        experiment_job = db.session.scalar(experiment_job_stmt)

        if experiment_job is None:
            raise EntityDoesNotExistError(
                RESOURCE_TYPE, job_id=job_id, experiment_id=experiment_id
            )

        return self._job_id_service.delete(
            job_id=job_id,
            log=log,
        )


class ExperimentJobIdStatusService(object):
    """The service methods for retrieving the status of a job by unique id."""

    @inject
    def __init__(
        self,
        experiment_job_id_service: ExperimentJobIdService,
    ) -> None:
        """Initialize the job status service.

        All arguments are provided via dependency injection.

        Args:
            experiment_job_id_service: An ExperimentJobIdService object.
        """
        self._experiment_job_id_service = experiment_job_id_service

    def get(
        self,
        experiment_id: int,
        job_id: int,
        **kwargs,
    ) -> dict[str, Any]:
        """Fetch a job's status by its unique id.

        Args:
            experiment_id: The unique id of the experiment.
            job_id: The unique id of the job.

        Returns:
            The status message job object if found, otherwise an error message.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get job status by id", job_id=job_id)

        job_dict = cast(
            utils.JobDict,
            self._experiment_job_id_service.get(
                experiment_id, job_id, error_if_not_found=True, log=log
            ),
        )
        job = job_dict["job"]

        return {"status": job.status, "id": job.resource_id}

    def modify(
        self,
        experiment_id: int,
        job_id: int,
        status: str,
        commit: bool = True,
        **kwargs,
    ) -> dict[str, Any] | None:
        """Modify a Job's status by unique id

        Args:
            experiment_id: The unique id of the experiment.
            job_id: The unique id of the job.
            status: The new status of the job.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The status message job object if found, otherwise an error message.
        """

        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Modify job status by id", job_id=job_id, status=status)

        job_dict = self._experiment_job_id_service.get(experiment_id, job_id, log=log)
        job = job_dict["job"]

        if status not in JOB_STATUS_TRANSITIONS.get(job.status, None):
            raise JobInvalidStatusTransitionError

        new_job = models.Job(
            timeout=job.timeout,
            status=status,
            description=job.description,
            resource=job.resource,
            creator=job.creator,
        )
        new_job.entry_point_job = job.entry_point_job
        new_job.experiment_job = job.experiment_job
        new_job.queue_job = job.queue_job

        db.session.add(new_job)

        if commit:
            db.session.commit()
            log.debug(
                "Job status modification successful",
                job_id=job_id,
                status=status,
            )

        return {"status": new_job.status, "id": job.resource_id}


class ExperimentJobIdMlflowrunService(object):
    """The service methods for managing the mlflow run id of a job by unique id."""

    @inject
    def __init__(
        self,
        experiment_job_id_service: ExperimentJobIdService,
    ) -> None:
        """Initialize the job status service.

        All arguments are provided via dependency injection.

        Args:
            experiment_job_id_service: An ExperimentJobIdService object.
        """
        self._experiment_job_id_service = experiment_job_id_service

    def get(
        self,
        experiment_id: int,
        job_id: int,
        **kwargs,
    ) -> dict[str, Any]:
        """Fetch a job's mlflow run id by its unique id.

        Args:
            experiment_id: The unique id of the experiment.
            job_id: The unique id of the job.

        Returns:
            The status message job object if found, otherwise an error message.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get job status by id", job_id=job_id)

        job_dict = cast(
            utils.JobDict,
            self._experiment_job_id_service.get(
                experiment_id, job_id, error_if_not_found=True, log=log
            ),
        )
        job = job_dict["job"]

        mlflow_run_id = (
            job.mlflow_run.mlflow_run_id if job.mlflow_run is not None else None
        )
        return {"mlflow_run_id": mlflow_run_id}

    def create(
        self,
        experiment_id: int,
        job_id: int,
        mlflow_run_id: str,
        commit: bool = True,
        **kwargs,
    ) -> dict[str, Any] | None:
        """Set a Job's mlflow run id by unique id

        Args:
            experiment_id: The unique id of the experiment.
            job_id: The unique id of the job.
            mlflow_run_id: The unique id of the mlflow run.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The status message job object if found, otherwise an error message.
        """

        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug(
            "Set the job's mlflow run id", job_id=job_id, mlflow_run_id=mlflow_run_id
        )

        job_dict = self._experiment_job_id_service.get(experiment_id, job_id, log=log)
        job = job_dict["job"]

        if job.mlflow_run is not None:
            raise JobMlflowRunAlreadySetError

        job.mlflow_run = models.JobMlflowRun(
            job_resource_id=job.resource_id,
            mlflow_run_id=mlflow_run_id,
        )

        if commit:
            db.session.commit()
            log.debug(
                "Setting Job mlflow run id successful",
                job_id=job_id,
                mlflow_run_id=mlflow_run_id,
            )

        return {"mlflow_run_id": job.mlflow_run.mlflow_run_id}


class JobIdMlflowrunService(object):
    """The service methods for managing the mlflow run id of a job by unique id."""

    @inject
    def __init__(
        self,
        job_id_service: JobIdService,
    ) -> None:
        """Initialize the job status service.

        All arguments are provided via dependency injection.

        Args:
            job_id_service: An JobIdService object.
        """
        self._job_id_service = job_id_service

    def get(
        self,
        job_id: int,
        **kwargs,
    ) -> dict[str, Any]:
        """Fetch a job's mlflow run id by its unique id.

        Args:
            job_id: The unique id of the job.

        Returns:
            The status message job object if found, otherwise an error message.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get job status by id", job_id=job_id)

        job_dict = cast(
            utils.JobDict,
            self._job_id_service.get(job_id, error_if_not_found=True, log=log),
        )
        job = job_dict["job"]

        mlflow_run_id = (
            job.mlflow_run.mlflow_run_id if job.mlflow_run is not None else None
        )
        return {"mlflow_run_id": mlflow_run_id}

    def create(
        self,
        job_id: int,
        mlflow_run_id: str,
        commit: bool = True,
        **kwargs,
    ) -> dict[str, Any] | None:
        """Set a Job's mlflow run id by unique id

        Args:
            job_id: The unique id of the job.
            mlflow_run_id: The unique id of the mlflow run.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The status message job object if found, otherwise an error message.
        """

        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug(
            "Set the job's mlflow run id", job_id=job_id, mlflow_run_id=mlflow_run_id
        )

        job_dict = cast(
            utils.JobDict,
            self._job_id_service.get(job_id, error_if_not_found=True, log=log),
        )
        job = job_dict["job"]

        if job.mlflow_run is not None:
            raise JobMlflowRunAlreadySetError

        job.mlflow_run = models.JobMlflowRun(
            job_resource_id=job.resource_id,
            mlflow_run_id=mlflow_run_id,
        )

        if commit:
            db.session.commit()
            log.debug(
                "Setting Job mlflow run id successful",
                job_id=job_id,
                mlflow_run_id=mlflow_run_id,
            )

        return {"mlflow_run_id": job.mlflow_run.mlflow_run_id}



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from pathlib import Path
from typing import Any, Iterator, Optional, Union

import structlog
from botocore.client import BaseClient

from dioptra.sdk.utilities.s3.uri import s3_uri_to_bucket_prefix


def _get_logger() -> Any:
    """
    Get a logger for this module.

    Returns:
        A logger object
    """
    return structlog.get_logger(__name__)


def get_s3_keys(
    s3: BaseClient, bucket: str, prefix: Optional[str] = None
) -> Iterator[str]:
    """
    Generate all keys in the given S3 bucket, having the given prefix.

    Args:
        s3: A boto3 S3 client object
        bucket: An S3 bucket name
        prefix: An optional key prefix

    Yields:
        Key names
    """

    resp = s3.list_objects_v2(Bucket=bucket, Prefix=prefix)

    for obj_info in resp.get("Contents", []):
        yield obj_info["Key"]

    while resp["IsTruncated"]:
        resp = s3.list_objects_v2(
            Bucket=bucket,
            Prefix=prefix,
            ContinuationToken=resp["NextContinuationToken"],
        )

        for obj_info in resp.get("Contents", []):
            yield obj_info["Key"]


def download_files_uri(
    s3: BaseClient, dest_dir: Path, s3_uri: str, preserve_key_paths: bool = False
):
    """
    Download all files from the given S3 URI to the given directory.

    Args:
        s3: A boto3 S3 client object
        dest_dir: A string or Path object referring to the directory where
            files will be downloaded to
        s3_uri: An S3 URI in the form "s3://<bucket>/<key_prefix>"
        preserve_key_paths: If True, append matching keys to dest_dir to obtain
            the final directory to download to.  The intermediate
            subdirectories are created automatically.  This causes the
            directory structure represented by the keys to be mirrored in the
            local filesystem.  If False, the last component of matching key
            paths are used as filenames and the rest of the paths are
            discarded.  This causes a directory structure in the bucket to be
            flattened to list of filenames.  No subdirectories are created in
            this case.
    """

    bucket, prefix = s3_uri_to_bucket_prefix(s3_uri)

    if not bucket:
        raise ValueError("S3 URIs must include a bucket: " + s3_uri)

    download_files(s3, dest_dir, bucket, prefix, preserve_key_paths)


def download_files(
    s3: BaseClient,
    dest_dir: Union[str, Path],
    bucket: str,
    prefix: Optional[str] = None,
    preserve_key_paths: bool = False,
):
    """
    Download all files from the given S3 bucket to the given directory, whose
    keys match the given prefix.

    Args:
        s3: A boto3 S3 client object
        dest_dir: A string or Path object referring to the directory where
            files will be downloaded to.  Must already exist.
        bucket: The name of an S3 bucket
        prefix: An optional S3 key prefix, used to limit what is downloaded
        preserve_key_paths: If True, append matching keys to dest_dir to obtain
            the final directory to download to.  The intermediate
            subdirectories are created automatically.  This causes the
            directory structure represented by the keys to be mirrored in the
            local filesystem.  If False, the last component of matching key
            paths are used as filenames and the rest of the paths are
            discarded.  This causes a directory structure in the bucket to be
            flattened to list of filenames.  No subdirectories are created in
            this case.
    """

    log = _get_logger()

    if isinstance(dest_dir, str):
        dest_dir = Path(dest_dir)

    for key in get_s3_keys(s3, bucket, prefix):
        if preserve_key_paths:
            key_path = dest_dir / key
            key_path.parent.mkdir(parents=True, exist_ok=True)
        else:
            key_path = dest_dir / Path(key).name

        log.debug("Downloading s3 key %s -> %s", key, str(key_path))
        s3.download_file(bucket, key, str(key_path))



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""
Try to derive better error messages for jsonschema validation errors.
"""

import collections
from typing import (
    Any,
    Callable,
    Iterable,
    MutableMapping,
    MutableSequence,
    Optional,
    Sequence,
    Union,
    cast,
)

import jsonschema
import jsonschema.exceptions
import jsonschema.protocols
import jsonschema.validators

# Type alias for general JSON schemas
_JsonSchema = Union[dict[str, Any], bool]

# Type alias for non-boolean JSON schemas
_JsonSchemaNonBool = dict[str, Any]

# Controls indentation of sub-parts of error messages
_INDENT_SIZE = 4


def _indent_lines(lines: MutableSequence[str]) -> MutableSequence[str]:
    """
    Add a level of indentation to each of the given lines.  The given
    line sequence is modified in-place, and for convenience also returned.

    Args:
        lines: A mutable sequence of line strings

    Returns:
        The same sequence object as was passed in, but containing indented
        lines.
    """
    for i, line in enumerate(lines):
        lines[i] = " " * _INDENT_SIZE + line

    return lines


def _schema_reference_to_path(ref: str) -> list[str]:
    """
    Convert a JSON-Schema reference to the same sort of path structure used in
    jsonschema ValidationError objects to identify locations in JSON.  This
    makes programmatically "following" the path to find what it refers to,
    much easier.  The general path structure is a list of strings/ints.  This
    implementation returns a list of strings and does not try to convert any
    path components to any other type.  Absent context, all-digits path
    components are ambiguous (keys or list indices?), so we mustn't change them.

    Args:
        ref: A reference value, as a fragment: "#" followed by a
            JSON-Pointer value.  If the fragment is "#" by itself, return an
            empty list.

    Returns:
        A path as a list of strings
    """

    if ref == "#":
        # I think "#" is legal, and refers to the whole document.
        ref_schema_path = []

    else:
        # Else, assume references start with "#/", i.e. are URL fragments
        # containing JSON pointers which start with "/".
        ref = ref[2:]  # strip the "#/"
        ref_schema_path = ref.split("/")

    return ref_schema_path


def _extract_schema_by_schema_path(
    schema_path: Iterable[Union[int, str]],
    full_schema: _JsonSchemaNonBool,
    schema: Optional[Union[dict[str, Any], list[Any]]] = None,
) -> Union[dict[str, Any], list[Any]]:
    """
    Find the schema sub-document referred to by a path.  The path must not
    include any "$ref" elements; references are transparently dereferenced as
    they are encountered.

    Args:
        schema_path: A path relative to "schema" (or if it is None, then
            relative to "full_schema"), as an iterable of strings/ints.
        full_schema: The full schema we are looking inside of.  This is used
            when traversing references, which are always absolute and therefore
            require restarting traversal from the top of the full schema.
        schema: The sub-part of full_schema being examined, or None.  If
            None, examine full_schema.

    Returns:
        The sub-part of schema referred to by the given path
    """

    if schema is None:
        schema = full_schema

    if isinstance(schema, dict) and "$ref" in schema:
        # jsonschema's schema paths don't actually contain "$ref" as an
        # element.  The paths pass through as if the referent was substituted
        # for the reference, and the reference wasn't even there.

        # the cast here is necessary: _schema_reference_to_path() is defined
        # to return list[str].  It never returns anything else, so I think it
        # would be incorrect to define it otherwise.  Mypy regards lists as
        # "invariant", i.e. list[A] and list[B] are considered incompatible
        # types no matter the relationship between A and B.  It is effectively
        # forcing the caller to never add anything other than strings to a
        # list[str] since that's what the called function is declared to
        # return.  In this case, the function may return me a list[str], but I
        # know I will be sole owner of it, and so I should be able to do
        # whatever I want with it, including adding values which are not
        # strings, and use it in contexts where non-string content is expected
        # (ints, in this case).
        ref_schema_path: MutableSequence[Union[int, str]] = cast(
            MutableSequence[Union[int, str]], _schema_reference_to_path(schema["$ref"])
        )
        # Here, schema_path may have integer list indices.  That's okay.
        ref_schema_path.extend(schema_path)

        result_schema = _extract_schema_by_schema_path(ref_schema_path, full_schema)

    else:
        schema_path_it = iter(schema_path)
        next_path_component = next(schema_path_it, None)

        if not next_path_component:
            result_schema = schema

        else:
            if isinstance(schema, list):
                # If current schema is a list, this path step must be
                # interpretable as an integer.  We won't actually know whether
                # a given path step which is a string comprised of all digits
                # refers to an all-digits property name, or a list index, until
                # this point.  This ambiguity occurs when splitting a JSON
                # pointer string to obtain a schema path.
                next_path_component = int(next_path_component)

            # next_path_component is correctly inferred to be Union[int, str],
            # but mypy does not consider that a valid index type.  Since
            # 'schema' can have different values at runtime (sometimes lists,
            # sometimes dicts), the below indexing can't always mean the same
            # thing: sometimes it's a key lookup in a dict, sometimes an index
            # lookup in a list.  As a static type checker, mypy seems to want
            # one meaning, and I couldn't figure out how to make that pass mypy
            # checks.
            subschema = schema[next_path_component]  # type: ignore

            result_schema = _extract_schema_by_schema_path(
                schema_path_it, full_schema, subschema
            )

    return result_schema


def _get_one_of_alternative_names(
    alternative_schemas: Iterable[Any], full_schema: _JsonSchemaNonBool
) -> list[str]:
    """
    Find names for the given alternative schemas.  The names are derived from
    "title" properties of the schemas.  Numeric suffixes may be introduced if
    necessary, to make them unique.

    Args:
        alternative_schemas: An iterable of sub-schemas, relative to
            full_schema
        full_schema: The full schema, of which the alternatives are a part.
            Important for being able to resolve references.

    Returns:
        A list of names
    """

    # It is possible, though unlikely, that more than one alternative has the
    # same name (title).  We will add a numeric counter suffix as necessary to
    # force alternative names to be unique.
    name_counts: MutableMapping[str, int] = collections.defaultdict(int)
    names = []

    for idx, alternative_schema in enumerate(alternative_schemas):
        if isinstance(alternative_schema, dict):
            # dereference if it's just a "$ref" schema
            alternative_schema = _extract_schema_by_schema_path(
                [], full_schema, alternative_schema
            )

            name = alternative_schema.get("title")
            if not name:
                name = "Alternative #" + str(idx + 1)

        else:
            # rare case... would only apply to true/false schemas I think.
            name = "Alternative #" + str(idx + 1)

        # uniquify names, just in case...
        name_count = name_counts[name]
        name_counts[name] += 1
        if name_count == 0:
            names.append(name)
        else:
            names.append("{}({})".format(name, name_count + 1))

    return names


def _is_valid_for_sub_schema(
    full_schema: _JsonSchemaNonBool, sub_schema: _JsonSchema, sub_instance: Any
) -> bool:
    """
    Run a validation of document sub_instance against sub_schema.

    Args:
        full_schema: The full schema, of which sub_schema is a part.
            Important for being able to resolve references.
        sub_schema: The schema to use for validation
        sub_instance: The instance document to validate

    Returns:
        True if sub_instance is valid; False if not
    """
    validator_class = jsonschema.validators.validator_for(full_schema)

    # Without this type annotation, the is_valid() call below is treated as
    # returning Any, and mypy errors since this function is defined to return
    # bool!  Even with the jsonschema type stubs, mypy gets confused.
    validator: jsonschema.protocols.Validator = validator_class(
        schema=sub_schema,
        # Need to construct a resolver from the full schema, since the
        # sub-schema might contain references relative to the full schema,
        # and we need to be able to resolve them.
        resolver=jsonschema.validators.RefResolver.from_schema(full_schema),
    )

    return validator.is_valid(sub_instance)


def _one_of_too_many_alternatives_satisfied_message_lines(
    error: jsonschema.exceptions.ValidationError, schema: _JsonSchemaNonBool
) -> list[str]:
    """
    Create an error message specifically about the situation where too many
    alternatives in a oneOf schema were valid.

    Args:
        error: The ValidationError object representing the aforementioned
            type of error
        schema: The schema whose validation failed

    Returns:
        An error message, as a list of lines (strings).  Returning a list
        of lines is convenient for callers, who may want to nest this message
        in another, with indented lines.
    """

    # In this specific case, validator_value contains the oneOf alternative
    # schemas, but the typing is such that mypy can't tell that's the case.
    alt_schemas = cast(Iterable[Any], error.validator_value)

    alt_names = _get_one_of_alternative_names(alt_schemas, schema)
    error_desc = "Must be exactly one of: {}".format(", ".join(alt_names))

    satisfied_alt_names = []
    for alt_name, alt_schema in zip(alt_names, alt_schemas):
        # Perform a little "mini" validation to determine which alternatives
        # were satisfied, and describe them in the error message.
        if _is_valid_for_sub_schema(schema, alt_schema, error.instance):
            satisfied_alt_names.append(alt_name)

    error_desc += ".  Content satisfied more than one alternative: {}.".format(
        ", ".join(satisfied_alt_names)
    )

    return [error_desc]


def _one_of_no_alternatives_satisfied_message_lines(
    error: jsonschema.exceptions.ValidationError,
    schema: _JsonSchemaNonBool,
    location_desc_callback: Callable[[Sequence[Union[int, str]]], str],
) -> list[str]:
    """
    Create an error message specifically about the situation where none of the
    alternatives in a oneOf schema were valid.

    Args:
        error: The ValidationError object representing the aforementioned
            type of error
        schema: The schema whose validation failed
        location_desc_callback: A callback function used to customize the
            description of the location of errors.  Takes a programmatic "path"
            structure as a sequence of strings/ints, and should return a nice
            one-line string description.

    Returns:
        An error message, as a list of lines (strings).  Returning a list
        of lines is convenient for callers, who may want to nest this message
        in another, with indented lines.
    """

    message_lines = []

    # In this specific case, validator_value contains the oneOf alternative
    # schemas, but the typing is such that mypy can't tell that's the case.
    alt_schemas = cast(Iterable[Any], error.validator_value)

    # First error message line describes the error in basic terms.

    alt_names = _get_one_of_alternative_names(alt_schemas, schema)
    basic_desc = (
        "Must be exactly one of: {}; all alternatives failed validation."
    ).format(", ".join(alt_names))

    message_lines.append(basic_desc)

    # Subsequent error lines give additional details about the error: errors
    # associated with each oneOf alternative sub-schema.

    # Maps an alternative name to a list of error objects associated with that
    # alternative.  We organize context errors according to the alternative
    # they apply to.
    errors_by_alt = collections.defaultdict(list)

    one_of_schema_path_len = len(error.absolute_schema_path)

    # required to assure mypy that error.context is non-null.  That is checked
    # before this function is called (in fact it is the exact criteria for the
    # call).  Otherwise, it is treated as a non-iterable Optional type.
    assert error.context
    for ctx_error in error.context:
        # schema paths for errors associated with the alternatives
        # will share a common prefix with the schema path for the
        # "oneOf" error.  The next element after the common portion
        # will be an int which is the index of the alternative.
        error_alt_idx = ctx_error.absolute_schema_path[one_of_schema_path_len]

        # Mypy infers Union[str, int] for error_alt_idx and complains that it
        # is not a valid index type.  As explained above, in this context, this
        # index must be an int.
        assert isinstance(error_alt_idx, int)
        errors_by_alt[alt_names[error_alt_idx]].append(ctx_error)

    for alt_name, alt_errors in errors_by_alt.items():
        message_lines.append(
            'Errors associated with alternative "{}":'.format(alt_name)
        )

        for alt_error in alt_errors:
            sub_message_lines = _validation_error_to_message_lines(
                alt_error, schema, location_desc_callback
            )
            message_lines.extend(_indent_lines(sub_message_lines))

    return message_lines


def _validation_error_to_message_lines(
    error: jsonschema.exceptions.ValidationError,
    schema: _JsonSchema,
    location_desc_callback: Callable[[Sequence[Union[int, str]]], str],
) -> list[str]:
    """
    Create a nice error message for the given error object.

    Args:
        error: A ValidationError object which represents some schema
            validation error
        schema: The schema whose validation failed
        location_desc_callback: A callback function used to customize the
            description of the location of errors.  Takes a programmatic "path"
            structure as a sequence of strings/ints, and should return a nice
            one-line string description.

    Returns:
        An error message, as a list of lines (strings).  Returning a list
        of lines is convenient for callers, who may want to nest this message
        in another, with indented lines.
    """

    # Describe "where" the error occurred
    location_desc = location_desc_callback(error.absolute_path)

    # Describe "what" error(s) occurred
    if error.validator == "oneOf":
        # If a "oneOf" error, the schema can't be a bool schema.  Bool schemas
        # have no subcomponents (it's just true or false), so they can't have a
        # "oneOf" validator.
        assert isinstance(schema, dict)

        if error.context:
            what_lines = _one_of_no_alternatives_satisfied_message_lines(
                error, schema, location_desc_callback
            )

        else:
            what_lines = _one_of_too_many_alternatives_satisfied_message_lines(
                error, schema
            )

    else:
        # fallback if we can't be more clever about our message
        what_lines = [error.message]

    if len(what_lines) == 1:
        message_lines = ["In {}: {}".format(location_desc, what_lines[0])]

    else:
        message_lines = ["In {}:".format(location_desc)]
        message_lines.extend(_indent_lines(what_lines))

    return message_lines


def json_path_to_string(path: Iterable[Any]) -> str:
    """
    Create a string representation of a JSON path as is used in jsonschema
    ValidationError objects.  For now, a filesystem-like syntax is used with
    slash-delimited strings, which I think winds up being the same as
    JSON-Pointer syntax (rfc6901).

    Args:
        path: A "path" into a JSON structure, as an iterable of values
            (strings and ints).

    Returns:
        A string representation of the path
    """
    # Use a filesystem-like syntax?
    return "/" + "/".join(str(elt) for elt in path)


def validation_error_to_message(
    error: jsonschema.exceptions.ValidationError,
    schema: _JsonSchema,
    location_desc_callback: Optional[Callable[[Sequence[Union[int, str]]], str]] = None,
) -> str:
    """
    Create a nice error message for the given error object.

    Args:
        error: A ValidationError object which represents some schema
            validation error
        schema: The schema whose validation failed
        location_desc_callback: A callback function used to customize the
            description of the location of errors.  Takes a programmatic "path"
            structure as a sequence of strings/ints, and should return a nice
            one-line string description.  Defaults to a simple generic
            implementation which produces descriptions which aren't very nice.

    Returns:
        An error message as a string
    """

    if location_desc_callback is None:
        location_desc_callback = json_path_to_string

    message_lines = _validation_error_to_message_lines(
        error, schema, location_desc_callback
    )

    message = "\n".join(message_lines)

    return message


def validation_errors_to_message(
    errors: Iterable[jsonschema.exceptions.ValidationError],
    schema: _JsonSchema,
    location_desc_callback: Optional[Callable[[Sequence[Union[int, str]]], str]] = None,
) -> str:
    """
    Create a nice error message for the given error objects.  This currently
    just creates error messages for each error individually, and then
    concatenates them all together with blank lines in between.

    Args:
        errors: An iterable of ValidationError objects which represent some
            schema validation errors
        schema: The schema whose validation failed
        location_desc_callback: A callback function used to customize the
            description of the location of errors.  Takes a programmatic "path"
            structure as a sequence of strings/ints, and should return a nice
            one-line string description.  Defaults to a simple generic
            implementation which produces descriptions which aren't very nice.

    Returns:
        An error message as a string.
    """
    messages = [
        validation_error_to_message(error, schema, location_desc_callback)
        for error in errors
    ]

    combined_message_str = "\n\n".join(messages)

    return combined_message_str



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
import pytest


@pytest.fixture
def context():
    return {
        "deployment_name": "Dioptra deployment",
        "__project_slug": "dioptra-deployment",
        "container_registry": "",
        "__container_namespace": "dioptra",
        "container_tag": "dev",
        "docker_compose_path": "/usr/local/bin/docker-compose",
        "systemd_required_mounts": "",
        "nginx_server_name": "dioptra.example.org",
        "nginx_expose_ports_on_localhost_only": "True",
        "postgres_container_tag": "15",
        "__restapi_http_port": "80",
        "__restapi_https_port": "443",
        "__db_admin_username": "postgres",
        "__db_admin_database": "postgres",
        "pgadmin_default_email": "dioptra@example.com",
        "num_tensorflow_cpu_workers": "1",
        "num_tensorflow_gpu_workers": "0",
        "num_pytorch_cpu_workers": "1",
        "num_pytorch_gpu_workers": "0",
        "__containers": {
            "networks": ["dioptra"],
            "nginx": {
                "image": "nginx",
                "namespace": "dioptra",
                "tag": "dev",
                "registry": "",
            },
            "mlflow_tracking": {
                "image": "mlflow-tracking",
                "namespace": "dioptra",
                "tag": "dev",
                "registry": "",
            },
            "restapi": {
                "image": "restapi",
                "namespace": "dioptra",
                "tag": "dev",
                "registry": "",
            },
            "tfcpu": {
                "image": "tensorflow2-cpu",
                "namespace": "dioptra",
                "tag": "dev",
                "registry": "",
            },
            "tfgpu": {
                "image": "tensorflow2-gpu",
                "namespace": "dioptra",
                "tag": "dev",
                "registry": "",
            },
            "pytorchcpu": {
                "image": "pytorch-cpu",
                "namespace": "dioptra",
                "tag": "dev",
                "registry": "",
            },
            "pytorchgpu": {
                "image": "pytorch-gpu",
                "namespace": "dioptra",
                "tag": "dev",
                "registry": "",
            },
            "argbash": {
                "image": "argbash",
                "namespace": "matejak",
                "tag": "latest",
                "registry": "",
            },
            "db": {
                "image": "postgres",
                "namespace": "",
                "tag": "{{ cookiecutter.postgres_container_tag }}",
                "registry": "",
            },
            "dbadmin": {
                "image": "pgadmin4",
                "namespace": "dpage",
                "tag": "latest",
                "registry": "",
            },
            "mc": {
                "image": "mc",
                "namespace": "minio",
                "tag": "latest",
                "registry": "",
            },
            "minio": {
                "image": "minio",
                "namespace": "minio",
                "tag": "latest",
                "registry": "",
            },
            "nodejs": {
                "image": "node",
                "namespace": "",
                "tag": "latest",
                "registry": "",
            },
            "redis": {
                "image": "redis",
                "namespace": "",
                "tag": "latest",
                "registry": "",
            },
        },
        "_is_pytest": "True",
    }


@pytest.fixture
def result(cookies, context):
    return cookies.bake(extra_context={**context})



"""Insert legacy into primary key constraint names.

Revision ID: b835738272a8
Revises: 8ffe4640e009
Create Date: 2024-04-07 17:12:23.966704

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "b835738272a8"
down_revision = "8ffe4640e009"
branch_labels = None
depends_on = None

POSTGRES_DIALECT = "postgresql"


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    dialect_name = op.get_context().dialect.name

    with op.batch_alter_table("legacy_experiments", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_legacy_experiments_name"))

    with op.batch_alter_table("legacy_jobs", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_legacy_jobs_status"))
        batch_op.drop_index(batch_op.f("ix_legacy_jobs_queue_id"))
        batch_op.drop_index(batch_op.f("ix_legacy_jobs_mlflow_run_id"))
        batch_op.drop_index(batch_op.f("ix_legacy_jobs_experiment_id"))
        batch_op.drop_constraint(
            batch_op.f("fk_legacy_jobs_status_legacy_job_statuses")
        )
        batch_op.drop_constraint(batch_op.f("fk_legacy_jobs_queue_id_legacy_queues"))
        batch_op.drop_constraint(
            batch_op.f("fk_legacy_jobs_experiment_id_legacy_experiments")
        )

    with op.batch_alter_table("legacy_queues", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_legacy_queues_name"))

    with op.batch_alter_table("legacy_queue_locks", schema=None) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("fk_legacy_queue_locks_queue_id_legacy_queues")
        )

    with op.batch_alter_table("legacy_experiments", schema=None) as batch_op:
        batch_op.drop_constraint(batch_op.f("pk_experiments"))

        if dialect_name == POSTGRES_DIALECT:
            batch_op.execute(
                'ALTER SEQUENCE IF EXISTS "experiments_experiment_id_seq" '
                'RENAME TO "legacy_experiments_experiment_id_seq"'
            )

        batch_op.create_primary_key("pk_legacy_experiments", ["experiment_id"])

    with op.batch_alter_table("legacy_job_statuses", schema=None) as batch_op:
        batch_op.drop_constraint(batch_op.f("pk_job_statuses"))
        batch_op.create_primary_key("pk_legacy_job_statuses", ["status"])

    with op.batch_alter_table("legacy_jobs", schema=None) as batch_op:
        batch_op.drop_constraint(batch_op.f("pk_jobs"))
        batch_op.create_primary_key("pk_legacy_jobs", ["job_id"])

    with op.batch_alter_table("legacy_queue_locks", schema=None) as batch_op:
        batch_op.drop_constraint(batch_op.f("pk_queue_locks"))
        batch_op.create_primary_key("pk_legacy_queue_locks", ["queue_id"])

    with op.batch_alter_table("legacy_queues", schema=None) as batch_op:
        batch_op.drop_constraint(batch_op.f("pk_queues"))

        if dialect_name == POSTGRES_DIALECT:
            batch_op.execute(
                'ALTER SEQUENCE IF EXISTS "queues_queue_id_seq" '
                'RENAME TO "legacy_queues_queue_id_seq"'
            )

        batch_op.create_primary_key("pk_legacy_queues", ["queue_id"])

    with op.batch_alter_table("legacy_users", schema=None) as batch_op:
        batch_op.drop_constraint(batch_op.f("pk_users"))

        if dialect_name == POSTGRES_DIALECT:
            batch_op.execute(
                'ALTER SEQUENCE IF EXISTS "users_user_id_seq" '
                'RENAME TO "legacy_users_user_id_seq"'
            )

        batch_op.create_primary_key("pk_legacy_users", ["user_id"])

    with op.batch_alter_table("legacy_experiments", schema=None) as batch_op:
        batch_op.create_index("ix_legacy_experiments_name", ["name"], unique=1)

    with op.batch_alter_table("legacy_jobs", schema=None) as batch_op:
        batch_op.create_index("ix_legacy_jobs_status", ["status"], unique=False)
        batch_op.create_index("ix_legacy_jobs_queue_id", ["queue_id"], unique=False)
        batch_op.create_index(
            "ix_legacy_jobs_mlflow_run_id", ["mlflow_run_id"], unique=False
        )
        batch_op.create_index(
            "ix_legacy_jobs_experiment_id", ["experiment_id"], unique=False
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_legacy_jobs_status_legacy_job_statuses"),
            "legacy_job_statuses",
            ["status"],
            ["status"],
        )
        batch_op.create_foreign_key(
            "fk_legacy_jobs_queue_id_legacy_queues",
            "legacy_queues",
            ["queue_id"],
            ["queue_id"],
        )
        batch_op.create_foreign_key(
            "fk_legacy_jobs_experiment_id_legacy_experiments",
            "legacy_experiments",
            ["experiment_id"],
            ["experiment_id"],
        )

    with op.batch_alter_table("legacy_queues", schema=None) as batch_op:
        batch_op.create_index("ix_legacy_queues_name", ["name"], unique=1)

    with op.batch_alter_table("legacy_queue_locks", schema=None) as batch_op:
        batch_op.create_foreign_key(
            "fk_legacy_queue_locks_queue_id_legacy_queues",
            "legacy_queues",
            ["queue_id"],
            ["queue_id"],
        )

    # Repeating this from Revision 32d5c0055e1b since it was retroactively added. Will
    # do nothing if types have already been dropped.
    if dialect_name == POSTGRES_DIALECT:
        op.execute("DROP TYPE IF EXISTS jobqueue; DROP TYPE IF EXISTS jobstatus;")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    dialect_name = op.get_context().dialect.name

    with op.batch_alter_table("legacy_experiments", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_legacy_experiments_name"))

    with op.batch_alter_table("legacy_jobs", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_legacy_jobs_status"))
        batch_op.drop_index(batch_op.f("ix_legacy_jobs_queue_id"))
        batch_op.drop_index(batch_op.f("ix_legacy_jobs_mlflow_run_id"))
        batch_op.drop_index(batch_op.f("ix_legacy_jobs_experiment_id"))
        batch_op.drop_constraint(
            batch_op.f("fk_legacy_jobs_status_legacy_job_statuses")
        )
        batch_op.drop_constraint(batch_op.f("fk_legacy_jobs_queue_id_legacy_queues"))
        batch_op.drop_constraint(
            batch_op.f("fk_legacy_jobs_experiment_id_legacy_experiments")
        )

    with op.batch_alter_table("legacy_queues", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_legacy_queues_name"))

    with op.batch_alter_table("legacy_queue_locks", schema=None) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("fk_legacy_queue_locks_queue_id_legacy_queues")
        )

    with op.batch_alter_table("legacy_experiments", schema=None) as batch_op:
        batch_op.drop_constraint(batch_op.f("pk_legacy_experiments"))

        if dialect_name == POSTGRES_DIALECT:
            batch_op.execute(
                'ALTER SEQUENCE IF EXISTS "legacy_experiments_experiment_id_seq" '
                'RENAME TO "experiments_experiment_id_seq"'
            )

        batch_op.create_primary_key("pk_experiments", ["experiment_id"])

    with op.batch_alter_table("legacy_job_statuses", schema=None) as batch_op:
        batch_op.drop_constraint(batch_op.f("pk_legacy_job_statuses"))
        batch_op.create_primary_key("pk_job_statuses", ["status"])

    with op.batch_alter_table("legacy_jobs", schema=None) as batch_op:
        batch_op.drop_constraint(batch_op.f("pk_legacy_jobs"))
        batch_op.create_primary_key("pk_jobs", ["job_id"])

    with op.batch_alter_table("legacy_queue_locks", schema=None) as batch_op:
        batch_op.drop_constraint(batch_op.f("pk_legacy_queue_locks"))
        batch_op.create_primary_key("pk_queue_locks", ["queue_id"])

    with op.batch_alter_table("legacy_queues", schema=None) as batch_op:
        batch_op.drop_constraint(batch_op.f("pk_legacy_queues"))

        if dialect_name == POSTGRES_DIALECT:
            batch_op.execute(
                'ALTER SEQUENCE IF EXISTS "legacy_queues_queue_id_seq" '
                'RENAME TO "queues_queue_id_seq"'
            )

        batch_op.create_primary_key("pk_queues", ["queue_id"])

    with op.batch_alter_table("legacy_users", schema=None) as batch_op:
        batch_op.drop_constraint(batch_op.f("pk_legacy_users"))

        if dialect_name == POSTGRES_DIALECT:
            batch_op.execute(
                'ALTER SEQUENCE IF EXISTS "legacy_users_user_id_seq" '
                'RENAME TO "users_user_id_seq"'
            )

        batch_op.create_primary_key("pk_users", ["user_id"])

    with op.batch_alter_table("legacy_experiments", schema=None) as batch_op:
        batch_op.create_index("ix_legacy_experiments_name", ["name"], unique=1)

    with op.batch_alter_table("legacy_jobs", schema=None) as batch_op:
        batch_op.create_index("ix_legacy_jobs_status", ["status"], unique=False)
        batch_op.create_index("ix_legacy_jobs_queue_id", ["queue_id"], unique=False)
        batch_op.create_index(
            "ix_legacy_jobs_mlflow_run_id", ["mlflow_run_id"], unique=False
        )
        batch_op.create_index(
            "ix_legacy_jobs_experiment_id", ["experiment_id"], unique=False
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_legacy_jobs_status_legacy_job_statuses"),
            "legacy_job_statuses",
            ["status"],
            ["status"],
        )
        batch_op.create_foreign_key(
            "fk_legacy_jobs_queue_id_legacy_queues",
            "legacy_queues",
            ["queue_id"],
            ["queue_id"],
        )
        batch_op.create_foreign_key(
            "fk_legacy_jobs_experiment_id_legacy_experiments",
            "legacy_experiments",
            ["experiment_id"],
            ["experiment_id"],
        )

    with op.batch_alter_table("legacy_queues", schema=None) as batch_op:
        batch_op.create_index("ix_legacy_queues_name", ["name"], unique=1)

    with op.batch_alter_table("legacy_queue_locks", schema=None) as batch_op:
        batch_op.create_foreign_key(
            "fk_legacy_queue_locks_queue_id_legacy_queues",
            "legacy_queues",
            ["queue_id"],
            ["queue_id"],
        )
    # ### end Alembic commands ###



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
import json
from pathlib import Path
from typing import Final

import structlog
from structlog.stdlib import BoundLogger

from dioptra.restapi.db import models

from .type_coercions import GlobalParameterType, coerce_to_type

LOGGER: BoundLogger = structlog.stdlib.get_logger()

JSON_FILE_ENCODING: Final[str] = "utf-8"


def export_job_parameters(
    job_param_values: list[models.EntryPointParameterValue],
    base_dir: Path,
    logger: BoundLogger | None = None,
) -> Path:
    """Export a job's parameters to a parameters.json file in a specified directory.

    Args:
        base_dir: The directory to export the parameters JSON file to.
        logger: A structlog logger object to use for logging. A new logger will be
            created if None.

    Returns:
        The path to the exported parameters.json file.
    """
    log = logger or LOGGER.new()  # noqa: F841
    job_params_json_path = Path(base_dir, "parameters").with_suffix(".json")

    job_parameters: dict[str, GlobalParameterType] = {}
    for param_value in job_param_values:
        value = coerce_to_type(
            x=param_value.value,
            type_name=param_value.parameter.parameter_type,
        )
        job_parameters[param_value.parameter.name] = value

    with job_params_json_path.open("wt", encoding=JSON_FILE_ENCODING) as f:
        json.dump(job_parameters, f, indent=2)

    return job_params_json_path



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from __future__ import annotations

import os
import shutil
import tarfile
from pathlib import Path

import pytest


@pytest.mark.parametrize(
    "filepath",
    ["tmp_unit_test.tar.gz"],
)
@pytest.mark.parametrize(
    ("dirs", "toremove", "filenames"),
    [
        (
            [
                "tmp_unit_test/test/dir/for/unit/tests",
                "tmp_unit_test/test/dir/for/unit/things",
                "tmp_unit_test/files/dir",
            ],
            "tmp_unit_test",
            ["test_random.py", "test_utils.py", "test_xyz.py"],
        ),
    ],
)
@pytest.mark.parametrize(
    "tarball_read_mode",
    ["r:gz"],
)
def test_extract_tarfile(
    filepath, dirs, toremove, filenames, tarball_read_mode
) -> None:
    from dioptra_builtins.artifacts.utils import extract_tarfile

    # generate files for tarfile
    filelist = []
    for m in dirs:
        for q in filenames:
            path_to_file = Path(m) / q
            filelist += [path_to_file]
            if not os.path.exists(m):
                os.makedirs(Path(m))
            f = open(path_to_file, "w")
            f.write("random text")
            f.close()

    for p in filelist:
        assert os.path.exists(p)

    # create temporary tarfile
    tar_file = tarfile.open(filepath, "w:gz")
    for z in filelist:
        tar_file.add(z)
    tar_file.close()

    # delete directories
    for dd in dirs:
        for root, ds, files in os.walk(dd):
            for f in files:
                os.unlink(os.path.join(root, f))
            for d in ds:
                shutil.rmtree(os.path.join(root, d))

    # check successful deletion
    for p in filelist:
        print(p)
        assert not os.path.exists(p)

    # check tarfile exists
    assert os.path.exists(filepath)

    # extract tarfile
    extract_tarfile(filepath, tarball_read_mode)

    # check tarfile extraction
    for p in filelist:
        assert os.path.exists(p)

    # clean up tarfile
    os.remove(filepath)

    # clean up directories
    for z in toremove:
        if os.path.exists(z):
            shutil.rmtree(z)


@pytest.mark.parametrize(
    ("dirs", "toremove"),
    [
        (
            [
                "tmp_unit_test/test/dir/for/unit/tests",
                "tmp_unit_test/test/dir/for/unit/things",
                "tmp_unit_test/files/dir",
            ],
            ["tmp_unit_test"],
        ),
        ([], []),
        (["tmp_unit_test_srcs"], ["tmp_unit_test_srcs"]),
    ],
)
def test_make_directories(dirs, toremove) -> None:
    from dioptra_builtins.artifacts.utils import make_directories

    for z in toremove:
        if os.path.exists(z):
            shutil.rmtree(z)
    assert not any([os.path.exists(d) for d in dirs])
    make_directories(dirs)
    assert all([os.path.isdir(d) for d in dirs])
    for z in toremove:
        if os.path.exists(z):
            shutil.rmtree(z)



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
import tarfile
import zipfile
from pathlib import Path
from tempfile import NamedTemporaryFile, TemporaryDirectory
from typing import IO

import structlog
from structlog.stdlib import BoundLogger

from dioptra.restapi.db import models

from ..schema import FileTypes
from .export_job_parameters import export_job_parameters
from .export_plugin_files import export_plugin_files
from .export_run_dioptra_job_script import export_run_dioptra_job_script
from .export_task_engine_yaml import export_task_engine_yaml

LOGGER: BoundLogger = structlog.stdlib.get_logger()


def package_job_files(
    job_id: int,
    experiment: models.Experiment,
    entry_point: models.EntryPoint,
    entry_point_plugin_files: list[models.EntryPointPluginFile],
    job_parameter_values: list[models.EntryPointParameterValue],
    plugin_parameter_types: list[models.PluginTaskParameterType],
    file_type: FileTypes,
    logger: BoundLogger | None = None,
) -> IO[bytes]:
    """Package the job files into a named temporary file.

    Args:
        job_id: The ID of the job to package the files for.
        experiment: The experiment the job is associated with.
        entry_point: The job's entrypoint.
        entry_point_plugin_files: The job's entrypoint plugin files.
        job_parameter_values: The job's assigned parameter values.
        plugin_parameter_types: The latest snapshots of the plugin parameter types
            accessible to the job.
        file_type: The type of file to package the job files into.
        logger: A structlog logger object to use for logging. A new logger will be
            created if None.

    Returns:
        The packaged job files as a named temporary file.
    """
    log = logger or LOGGER.new()  # noqa: F841

    with TemporaryDirectory() as tmp_dir:
        base_dir = Path(tmp_dir)
        plugin_base_dir = base_dir / "plugins"

        _ = export_plugin_files(
            entry_point_plugin_files=entry_point_plugin_files,
            plugins_base_dir=plugin_base_dir,
            logger=log,
        )
        task_engine_yaml_path = export_task_engine_yaml(
            entrypoint=entry_point,
            entry_point_plugin_files=entry_point_plugin_files,
            plugin_parameter_types=plugin_parameter_types,
            base_dir=base_dir,
            logger=log,
        )
        job_params_json_path = export_job_parameters(
            job_param_values=job_parameter_values, base_dir=base_dir, logger=log
        )
        run_dioptra_job_script = export_run_dioptra_job_script(
            job_id=job_id,
            experiment_id=experiment.resource_id,
            task_engine_yaml_path=task_engine_yaml_path.name,
            job_params_json_path=job_params_json_path.name,
            base_dir=base_dir,
            logger=log,
        )

        if file_type == FileTypes.TAR_GZ:
            return _package_as_tarfile(
                plugin_base_dir=plugin_base_dir,
                task_engine_yaml_path=task_engine_yaml_path,
                job_params_json_path=job_params_json_path,
                script_path=run_dioptra_job_script,
                logger=log,
            )

        if file_type == FileTypes.ZIP:
            return _package_as_zipfile(
                plugin_base_dir=plugin_base_dir,
                task_engine_yaml_path=task_engine_yaml_path,
                job_params_json_path=job_params_json_path,
                script_path=run_dioptra_job_script,
                logger=log,
            )


def _package_as_tarfile(
    plugin_base_dir: Path,
    task_engine_yaml_path: Path,
    job_params_json_path: Path,
    script_path: Path,
    logger: BoundLogger | None = None,
) -> IO[bytes]:
    log = logger or LOGGER.new()  # noqa: F841

    package_file = NamedTemporaryFile(suffix=".tar.gz")
    with tarfile.open(fileobj=package_file, mode="w:gz") as tar:
        tar.add(str(plugin_base_dir), plugin_base_dir.name, recursive=True)
        tar.add(str(task_engine_yaml_path), task_engine_yaml_path.name, recursive=False)
        tar.add(str(job_params_json_path), job_params_json_path.name, recursive=False)
        tar.add(str(script_path), script_path.name, recursive=False)

    package_file.seek(0)
    return package_file


def _package_as_zipfile(
    plugin_base_dir: Path,
    task_engine_yaml_path: Path,
    job_params_json_path: Path,
    script_path: Path,
    logger: BoundLogger | None = None,
) -> IO[bytes]:
    log = logger or LOGGER.new()  # noqa: F841

    package_file = NamedTemporaryFile(suffix=".zip")
    with zipfile.ZipFile(package_file, "w") as zipf:
        for file in plugin_base_dir.rglob("*"):
            zipf.write(file, file.relative_to(plugin_base_dir))

        zipf.write(task_engine_yaml_path, task_engine_yaml_path.name)
        zipf.write(job_params_json_path, job_params_json_path.name)
        zipf.write(script_path, script_path.name)

    package_file.seek(0)
    return package_file



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from __future__ import annotations

import tarfile
from pathlib import Path
from tarfile import TarFile, TarInfo
from typing import IO, List, Optional, Union

import structlog
from structlog.stdlib import BoundLogger

LOGGER: BoundLogger = structlog.stdlib.get_logger()


class UnsafeArchiveMemberPath(Exception):
    """
    Instances represent unsafe paths in an archive file.  E.g. those which
    contain "..", which might cause files to be written outside a target
    directory.
    """

    def __init__(self, archive_member_path: Union[str, Path]):
        message = "Unsafe archive member path: " + str(archive_member_path)
        super().__init__(message)

        self.archive_member_path = Path(archive_member_path)


class IOFileService(object):
    def safe_extract_archive(
        self,
        output_dir: Union[str, Path],
        archive_file_path: Optional[Union[str, Path]] = None,
        archive_fileobj: Optional[IO[bytes]] = None,
        preserve_paths: bool = False,
        **kwargs,
    ) -> List[str]:
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        extracted_files: List[str] = []

        with self._tarfile_open(
            archive_file_path, archive_fileobj, log=log
        ) as f_archive:
            for archive_file_info in f_archive:
                safe_file_path: Optional[Path] = self.sanitize_file_path(
                    filepath=archive_file_info.name,
                    path_prefix=output_dir,
                    preserve_paths=preserve_paths,
                    log=log,
                )

                if safe_file_path is None:
                    raise UnsafeArchiveMemberPath(archive_file_info.name)

                response: Optional[str] = self.safe_extract_archive_file(
                    file_path=safe_file_path,
                    f_archive=f_archive,
                    archive_file_info=archive_file_info,
                    log=log,
                )

                if response is not None:
                    extracted_files.append(response)

        return extracted_files

    @staticmethod
    def safe_extract_archive_file(
        file_path: Path,
        f_archive: TarFile,
        archive_file_info: TarInfo,
        **kwargs,
    ) -> Optional[str]:
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        if not archive_file_info.isfile():
            return None

        f_archive_file = f_archive.extractfile(archive_file_info)

        if f_archive_file is None:
            return None

        file_path.parent.mkdir(parents=True, exist_ok=True)

        with file_path.open("wb") as f:
            f.write(f_archive_file.read())

        log.debug("File extracted from archive", extracted_file=str(file_path))

        return str(file_path)

    @staticmethod
    def sanitize_file_path(
        filepath: Union[str, Path],
        path_prefix: Optional[Union[str, Path]] = None,
        preserve_paths: bool = False,
        **kwargs,
    ) -> Optional[Path]:
        log: BoundLogger = kwargs.get("log", LOGGER.new())  # noqa: F841

        filepath = Path(filepath)

        if preserve_paths:
            # Let's say we are willing to drop drive letters and root
            # directory.  But any ".." path component is considered
            # unsafe.  Also, if the path is only an anchor, it is unsafe.
            filepath_parts = filepath.parts

            if filepath.anchor:
                # The "anchor" is a combination of windows drive letter
                # (if any), and root path (if any).  Part 0 always contains the
                # anchor, if present.
                filepath_parts = filepath_parts[1:]

            safe_filepath: Optional[Union[str, Path]]
            if len(filepath_parts) == 0 or ".." in filepath_parts:
                safe_filepath = None
            else:
                safe_filepath = Path(*filepath_parts)

        else:
            safe_filepath = filepath.name

        if safe_filepath is None:
            safe_path = None
        else:
            safe_path = Path(path_prefix or Path.cwd()) / safe_filepath

        return safe_path

    @staticmethod
    def _tarfile_open(
        file_path: Optional[Union[str, Path]] = None,
        fileobj: Optional[IO[bytes]] = None,
        **kwargs,
    ) -> TarFile:
        log: BoundLogger = kwargs.get("log", LOGGER.new())  # noqa: F841

        def open_file_path() -> TarFile:
            return tarfile.open(name=file_path, mode="r:*")

        def open_fileobj() -> TarFile:
            return tarfile.open(fileobj=fileobj)

        if fileobj is not None:
            return open_fileobj()

        return open_file_path()



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from __future__ import annotations

from typing import Optional, Tuple

import structlog
from structlog.stdlib import BoundLogger

LOGGER: BoundLogger = structlog.stdlib.get_logger()

try:
    import tensorflow as tf
    from tensorflow.keras import Model
    from tensorflow.keras.applications.efficientnet import (
        EfficientNetB0,
        EfficientNetB1,
        EfficientNetB2,
        EfficientNetB3,
        EfficientNetB4,
        EfficientNetB5,
        EfficientNetB6,
        EfficientNetB7,
    )
    from tensorflow.keras.applications.efficientnet import (
        preprocess_input as efficient_net_preprocess_input,
    )
    from tensorflow.keras.applications.mobilenet_v2 import MobileNetV2
    from tensorflow.keras.applications.mobilenet_v2 import (
        preprocess_input as mobilenet_v2_preprocess_input,
    )
    from tensorflow.keras.layers import InputLayer

except ImportError:  # pragma: nocover
    LOGGER.warn(
        "Unable to import one or more optional packages, functionality may be reduced",
        package="tensorflow",
    )


class EfficientNetBackbone(Model):
    def __init__(
        self,
        flavor: str = "B4",
        input_shape: Optional[Tuple[int, int, int]] = None,
        weights: str = "imagenet",
        input_tensor: Optional[InputLayer] = None,
        pooling: Optional[str] = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        efficient_net_registry = dict(
            b0=EfficientNetB0,
            b1=EfficientNetB1,
            b2=EfficientNetB2,
            b3=EfficientNetB3,
            b4=EfficientNetB4,
            b5=EfficientNetB5,
            b6=EfficientNetB6,
            b7=EfficientNetB7,
        )
        self.model = efficient_net_registry[flavor.strip().lower()](
            include_top=False,
            weights=weights,
            input_tensor=input_tensor,
            input_shape=input_shape,
            pooling=pooling,
        )
        self.model.trainable = False

    @property
    def output_grid_shape(self):
        return (self.model.output_shape[1], self.model.output_shape[2])

    def call(self, inputs):
        x = self.normalize(inputs)
        return self.model(x, training=False)

    @staticmethod
    def normalize(inputs):
        x = tf.cast(inputs, tf.float32)
        return efficient_net_preprocess_input(x)


class MobileNetV2Backbone(Model):
    def __init__(
        self,
        input_shape: Optional[Tuple[int, int, int]] = None,
        alpha: float = 1.0,
        weights: str = "imagenet",
        input_tensor: Optional[InputLayer] = None,
        pooling: Optional[str] = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.model = MobileNetV2(
            input_shape=input_shape,
            alpha=alpha,
            include_top=False,
            weights=weights,
            input_tensor=input_tensor,
            pooling=pooling,
        )
        self.model.trainable = False

    @property
    def output_grid_shape(self):
        return (self.model.output_shape[1], self.model.output_shape[2])

    def call(self, inputs):
        x = self.normalize(inputs)
        return self.model(x, training=False)

    @staticmethod
    def normalize(inputs):
        x = tf.cast(inputs, tf.float32)
        return mobilenet_v2_preprocess_input(x)



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The server-side functions that perform snapshots sub endpoint operations."""
from __future__ import annotations

from typing import Any, Type

import structlog
from injector import inject
from sqlalchemy import func, select
from structlog.stdlib import BoundLogger

from dioptra.restapi.db import db, models
from dioptra.restapi.errors import BackendDatabaseError, EntityDoesNotExistError
from dioptra.restapi.v1.shared.search_parser import construct_sql_query_filters

LOGGER: BoundLogger = structlog.stdlib.get_logger()


class ResourceSnapshotsService(object):
    @inject
    def __init__(
        self,
        resource_model: Type[models.ResourceSnapshot],
        resource_type: str,
        searchable_fields: dict[str, Any],
    ) -> None:
        """Initialize the draft service.

        Args:
            resource_model:
            resource_type:
            searchable_fields:
        """
        self.resource_model = resource_model
        self._resource_type = resource_type
        self._searchable_fields = searchable_fields

    def get(
        self,
        resource_id: int,
        search_string: str,
        page_index: int,
        page_length: int,
        error_if_not_found: bool = False,
        **kwargs,
    ) -> tuple[list[dict[str, Any]], int] | None:
        """Fetch a list of snapshots of a resource.

        Args:
            resource_id: The unique id of the resource.
            search_string: A search string used to filter results.
            page_index: The index of the first snapshot to be returned.
            page_length: The maximum number of snapshots to be returned.
            error_if_not_found: If True, raise an error if the resource is not found.
                Defaults to False.

        Returns:
            The list of resource snapshots of the resource object if found, otherwise
                None.

        Raises:
            EntityDoesNotExistError: If the resource is not found and
                `error_if_not_found` is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get resource snapshots by id", resource_id=resource_id)

        stmt = select(models.Resource).filter_by(
            resource_id=resource_id, resource_type=self._resource_type, is_deleted=False
        )
        resource = db.session.scalars(stmt).first()

        if resource is None:
            if error_if_not_found:
                raise EntityDoesNotExistError(
                    self._resource_type, resource_id=resource_id
                )

            return None

        filters = list()

        if search_string:
            filters.append(
                construct_sql_query_filters(search_string, self._searchable_fields)
            )

        stmt = (
            select(func.count(self.resource_model.resource_id))  # type: ignore
            .join(models.Resource)
            .where(
                *filters,
                models.Resource.resource_id == resource_id,
                models.Resource.is_deleted == False,  # noqa: E712
            )
        )
        total_num_snapshots = db.session.scalars(stmt).unique().first()

        if total_num_snapshots is None:
            log.error(
                "The database query returned a None when counting the number of "
                "snapshots when it should return a number.",
                sql=str(stmt),
            )
            raise BackendDatabaseError

        if total_num_snapshots == 0:
            return [], total_num_snapshots

        stmt = (
            select(self.resource_model)
            .join(models.Resource)
            .where(
                *filters,
                models.Resource.resource_id == resource_id,
                models.Resource.is_deleted == False,  # noqa: E712
            )
            .order_by(self.resource_model.created_on)
            .offset(page_index)
            .limit(page_length)
        )
        snapshots = [
            {self._resource_type: snapshot, "has_draft": None}
            for snapshot in db.session.scalars(stmt).unique()
        ]

        return snapshots, total_num_snapshots


class ResourceSnapshotsIdService(object):
    @inject
    def __init__(
        self,
        resource_model: Type[models.ResourceSnapshot],
        resource_type: str,
    ) -> None:
        """Initialize the draft service.

        Args:
            resource_model: model
            resource_type: type
        """
        self.resource_model = resource_model
        self._resource_type = resource_type

    def get(
        self,
        resource_id: int,
        snapshot_id: int,
        error_if_not_found: bool = False,
        **kwargs,
    ) -> dict[str, Any] | None:
        """Fetch a specific snapshot of a resource.

        Args:
            resource_id: The unique id of the resource.
            snapshot_id: A search string used to filter results.
            error_if_not_found: If True, raise an error if the resource is not found.
                Defaults to False.

        Returns:
            The requested snapshot the resource object if found, otherwise None.

        Raises:
            EntityDoesNotExistError: If the resource is not found and
                `error_if_not_found` is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get resource snapshot by id", resource_id=resource_id)

        stmt = select(models.Resource).filter_by(
            resource_id=resource_id, resource_type=self._resource_type, is_deleted=False
        )
        resource = db.session.scalars(stmt).first()

        if resource is None:
            if error_if_not_found:
                raise EntityDoesNotExistError(
                    self._resource_type, resource_id=resource_id
                )

            return None

        stmt = (
            select(self.resource_model)
            .join(models.Resource)
            .where(
                models.ResourceSnapshot.resource_snapshot_id == snapshot_id,
                models.Resource.resource_id == resource_id,
                models.Resource.is_deleted == False,  # noqa: E712
            )
        )
        snapshot = db.session.scalars(stmt).first()

        if snapshot is None:
            if error_if_not_found:
                raise EntityDoesNotExistError(
                    self._resource_type + "_snapshot", snapshot_id=snapshot_id
                )

            return None

        return {self._resource_type: snapshot, "has_draft": None}



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from __future__ import annotations

from typing import Optional, Tuple

import numpy as np
import structlog
from structlog.stdlib import BoundLogger

from .augmentations import ObjectDetectionAugmentations

LOGGER: BoundLogger = structlog.stdlib.get_logger()

try:
    import imgaug.augmenters as iaa
    from imgaug.augmentables.bbs import BoundingBox, BoundingBoxesOnImage

except ImportError:  # pragma: nocover
    LOGGER.warn(
        "Unable to import one or more optional packages, functionality may be reduced",
        package="imgaug",
    )


class ImgAugObjectDetectionAugmentations(ObjectDetectionAugmentations):
    def __init__(
        self, image_dimensions: Tuple[int, int], augmenters: iaa.Sequential
    ) -> None:
        self._image_dimensions = image_dimensions
        self._augmenters = augmenters

    @classmethod
    def use_minimal_augmenters(
        cls, image_dimensions: Tuple[int, int], seed: Optional[int] = None
    ) -> ImgAugObjectDetectionAugmentations:
        augmenters: iaa.Sequential = iaa.Sequential(
            [
                iaa.Fliplr(0.5),
                iaa.Flipud(0.2),
                iaa.MultiplyBrightness((0.8, 1.2)),
                iaa.LinearContrast((0.75, 1.5)),
                iaa.Multiply((0.8, 1.2)),
            ],
            random_order=True,
            seed=seed,
        )

        return cls(
            image_dimensions=image_dimensions,
            augmenters=augmenters,
        )

    @classmethod
    def use_light_augmenters(
        cls, image_dimensions: Tuple[int, int], seed: Optional[int] = None
    ) -> ImgAugObjectDetectionAugmentations:
        augmenters: iaa.Sequential = iaa.Sequential(
            [
                iaa.Fliplr(0.5),
                iaa.Flipud(0.2),
                iaa.Crop(percent=(0, 0.1)),
                iaa.Sometimes(0.5, iaa.GaussianBlur(sigma=(0, 0.5))),
                iaa.LinearContrast((0.75, 1.5), per_channel=0.2),
                iaa.MultiplyBrightness((0.75, 1.5)),
                iaa.AdditiveGaussianNoise(
                    loc=0, scale=(0.0, 0.05 * 255), per_channel=0.5
                ),
                iaa.Multiply((0.8, 1.2), per_channel=0.2),
                iaa.Affine(
                    scale={"x": (0.8, 1.2), "y": (0.8, 1.2)},
                    translate_percent={"x": (-0.2, 0.2), "y": (-0.2, 0.2)},
                    rotate=(-25, 25),
                    shear=(-8, 8),
                ),
            ],
            random_order=True,
            seed=seed,
        )

        return cls(
            image_dimensions=image_dimensions,
            augmenters=augmenters,
        )

    @classmethod
    def use_heavy_augmenters(
        cls, image_dimensions: Tuple[int, int], seed: Optional[int] = None
    ) -> ImgAugObjectDetectionAugmentations:
        augmenters: iaa.Sequential = iaa.Sequential(
            [
                iaa.Fliplr(0.5),
                iaa.Flipud(0.2),
                iaa.Sometimes(0.5, iaa.Crop(percent=(0, 0.1))),
                iaa.Sometimes(
                    0.5,
                    iaa.Affine(
                        scale={"x": (0.8, 1.2), "y": (0.8, 1.2)},
                        translate_percent={"x": (-0.2, 0.2), "y": (-0.2, 0.2)},
                        rotate=(-45, 45),
                        shear=(-16, 16),
                    ),
                ),
                iaa.Sometimes(0.5, iaa.GaussianBlur(sigma=(0, 2.0))),
                iaa.Sometimes(
                    0.5,
                    iaa.AdditiveGaussianNoise(
                        loc=0, scale=(0.0, 0.05 * 255), per_channel=0.5
                    ),
                ),
                iaa.Sometimes(
                    0.5,
                    iaa.OneOf(
                        [
                            iaa.Dropout((0.01, 0.1), per_channel=0.5),
                            iaa.CoarseDropout(
                                (0.02, 0.1), size_percent=(0.01, 0.05), per_channel=0.2
                            ),
                        ]
                    ),
                ),
                iaa.Sometimes(0.5, iaa.Multiply((0.5, 1.5), per_channel=0.5)),
                iaa.Sometimes(0.5, iaa.MultiplyBrightness((0.75, 1.5))),
                iaa.Sometimes(0.5, iaa.LinearContrast((0.5, 2.0), per_channel=0.5)),
                iaa.Sometimes(0.5, iaa.Grayscale(alpha=(0.0, 1.0))),
            ],
            random_order=True,
            seed=seed,
        )

        return cls(
            image_dimensions=image_dimensions,
            augmenters=augmenters,
        )

    @property
    def augmenters(self) -> iaa.Sequential:
        return self._augmenters

    @property
    def image_height(self) -> int:
        return self._image_dimensions[0]

    @property
    def image_width(self) -> int:
        return self._image_dimensions[1]

    def augment(
        self, image: np.ndarray, bounding_boxes: np.ndarray, labels: np.ndarray
    ):
        augmented_bboxes: list[tuple[int, int, int, int]] = []
        bboxes_on_image = self._as_bounding_boxes_on_image(
            bounding_boxes, labels=labels
        )

        # Ensure at least one bounding box within image boundaries after augmentation
        while len(augmented_bboxes) == 0:
            augmented_image, augmented_bboxes_on_image = self.augmenters(
                image=image.astype("uint8"),
                bounding_boxes=bboxes_on_image,
            )
            augmented_bboxes_pruned = (
                augmented_bboxes_on_image.remove_out_of_image().clip_out_of_image()
            )
            augmented_bboxes = [
                (
                    x.x1_int / self.image_width,
                    x.y1_int / self.image_height,
                    x.x2_int / self.image_width,
                    x.y2_int / self.image_height,
                )
                for x in augmented_bboxes_pruned
            ]
            labels_pruned = [int(x.label) for x in augmented_bboxes_pruned]

        return (
            augmented_image.astype("float32"),
            np.array(augmented_bboxes, dtype="float32"),
            np.array(labels_pruned, dtype="int32"),
        )

    def _as_bounding_boxes_on_image(
        self, corner_bboxes: np.ndarray, labels: np.ndarray
    ) -> BoundingBoxesOnImage:
        return BoundingBoxesOnImage(
            [
                BoundingBox(
                    x1=bbox[0] * self.image_width,
                    y1=bbox[1] * self.image_height,
                    x2=bbox[2] * self.image_width,
                    y2=bbox[3] * self.image_height,
                    label=str(label),
                )
                for bbox, label in zip(
                    corner_bboxes.tolist(),
                    labels.tolist(),
                )
            ],
            shape=(self.image_height, self.image_width),
        )



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The module defining the endpoints for PluginParameterType resources."""
import uuid
from typing import cast

import structlog
from flask import request
from flask_accepts import accepts, responds
from flask_login import login_required
from flask_restx import Namespace, Resource
from injector import inject
from structlog.stdlib import BoundLogger

from dioptra.restapi.db import models
from dioptra.restapi.routes import V1_PLUGIN_PARAMETER_TYPES_ROUTE
from dioptra.restapi.v1 import utils
from dioptra.restapi.v1.schemas import IdStatusResponseSchema
from dioptra.restapi.v1.shared.drafts.controller import (
    generate_resource_drafts_endpoint,
    generate_resource_drafts_id_endpoint,
    generate_resource_id_draft_endpoint,
)
from dioptra.restapi.v1.shared.snapshots.controller import (
    generate_resource_snapshots_endpoint,
    generate_resource_snapshots_id_endpoint,
)
from dioptra.restapi.v1.shared.tags.controller import (
    generate_resource_tags_endpoint,
    generate_resource_tags_id_endpoint,
)

from .schema import (
    PluginParameterTypeGetQueryParameters,
    PluginParameterTypeMutableFieldsSchema,
    PluginParameterTypePageSchema,
    PluginParameterTypeSchema,
)
from .service import (
    RESOURCE_TYPE,
    SEARCHABLE_FIELDS,
    PluginParameterTypeIdService,
    PluginParameterTypeService,
)

LOGGER: BoundLogger = structlog.stdlib.get_logger()

api: Namespace = Namespace(
    "PluginParameterTypes", description="Plugin Parameter Types endpoint"
)


@api.route("/")
class PluginParameterTypeEndpoint(Resource):
    @inject
    def __init__(
        self, plugin_parameter_type_service: PluginParameterTypeService, *args, **kwargs
    ) -> None:
        """Initialize the plugin parameter type resource.

        All arguments are provided via dependency injection.

        Args:
            plugin_parameter_type_service: A PluginParameterTypeService object.
        """
        self._plugin_parameter_type_service = plugin_parameter_type_service
        super().__init__(*args, **kwargs)

    @login_required
    @accepts(query_params_schema=PluginParameterTypeGetQueryParameters, api=api)
    @responds(schema=PluginParameterTypePageSchema, api=api)
    def get(self):
        """Gets a list of all PluginParameterType resources."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="PluginParameterType",
            request_type="GET",
        )
        parsed_query_params = request.parsed_query_params  # noqa: F841

        group_id = parsed_query_params["group_id"]
        search_string = parsed_query_params["search"]
        page_index = parsed_query_params["index"]
        page_length = parsed_query_params["page_length"]
        sort_by_string = parsed_query_params["sort_by"]
        descending = parsed_query_params["descending"]

        (
            plugin_parameter_types,
            total_num_plugin_param_types,
        ) = self._plugin_parameter_type_service.get(
            group_id=group_id,
            search_string=search_string,
            page_index=page_index,
            page_length=page_length,
            sort_by_string=sort_by_string,
            descending=descending,
            log=log,
        )
        return utils.build_paging_envelope(
            V1_PLUGIN_PARAMETER_TYPES_ROUTE,
            build_fn=utils.build_plugin_parameter_type,
            data=plugin_parameter_types,
            group_id=group_id,
            query=search_string,
            draft_type=None,
            index=page_index,
            length=page_length,
            total_num_elements=total_num_plugin_param_types,
            sort_by=sort_by_string,
            descending=descending,
        )

    @login_required
    @accepts(schema=PluginParameterTypeSchema, api=api)
    @responds(schema=PluginParameterTypeSchema, api=api)
    def post(self):
        """Creates a PluginParameterType resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="PluginParameterType",
            request_type="POST",
        )
        parsed_obj = request.parsed_obj  # noqa: F841

        plugin_parameter_type = self._plugin_parameter_type_service.create(
            name=parsed_obj["name"],
            structure=parsed_obj["structure"],
            description=parsed_obj["description"],
            group_id=parsed_obj["group_id"],
            log=log,
        )
        return utils.build_plugin_parameter_type(plugin_parameter_type)


@api.route("/<int:id>")
@api.param("id", "ID for the PluginParameterType resource.")
class PluginParameterTypeIdEndpoint(Resource):
    @inject
    def __init__(
        self,
        plugin_parameter_type_id_service: PluginParameterTypeIdService,
        *args,
        **kwargs,
    ) -> None:
        """Initialize the plugin parameter type resource.

        All arguments are provided via dependency injection.

        Args:
            plugin_parameter_type_id_service: A PluginParameterTypeIdService object.
        """
        self._plugin_parameter_type_id_service = plugin_parameter_type_id_service
        super().__init__(*args, **kwargs)

    @login_required
    @responds(schema=PluginParameterTypeSchema, api=api)
    def get(self, id: int):
        """Gets a PluginParameterType resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="PluginParameterType",
            request_type="GET",
            id=id,
        )
        plugin_parameter_type = cast(
            models.PluginTaskParameterType,
            self._plugin_parameter_type_id_service.get(
                id, error_if_not_found=True, log=log
            ),
        )
        return utils.build_plugin_parameter_type(plugin_parameter_type)

    @login_required
    @responds(schema=IdStatusResponseSchema, api=api)
    def delete(self, id: int):
        """Deletes a PluginParameterType resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="PluginParameterType",
            request_type="DELETE",
            id=id,
        )
        return self._plugin_parameter_type_id_service.delete(
            plugin_parameter_type_id=id, log=log
        )

    @login_required
    @accepts(schema=PluginParameterTypeMutableFieldsSchema, api=api)
    @responds(schema=PluginParameterTypeSchema, api=api)
    def put(self, id: int):
        """Modifies a PluginParameterType resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="PluginParameterType",
            request_type="PUT",
            id=id,
        )
        parsed_obj = request.parsed_obj  # type: ignore
        plugin_parameter_type = cast(
            models.PluginTaskParameterType,
            self._plugin_parameter_type_id_service.modify(
                id,
                name=parsed_obj["name"],
                structure=parsed_obj["structure"],
                description=parsed_obj["description"],
                error_if_not_found=True,
                log=log,
            ),
        )
        return utils.build_plugin_parameter_type(plugin_parameter_type)


PluginParameterTypeDraftResource = generate_resource_drafts_endpoint(
    api=api,
    resource_name=RESOURCE_TYPE,
    route_prefix=V1_PLUGIN_PARAMETER_TYPES_ROUTE,
    request_schema=PluginParameterTypeSchema,
)

PluginParameterTypeDraftIdResource = generate_resource_drafts_id_endpoint(
    api=api,
    resource_name=RESOURCE_TYPE,
    request_schema=PluginParameterTypeMutableFieldsSchema,
)

PluginParameterTypeIdDraftIdResource = generate_resource_id_draft_endpoint(
    api=api,
    resource_name=RESOURCE_TYPE,
    request_schema=PluginParameterTypeMutableFieldsSchema,
)

PluginParameterTypeSnapshotsResource = generate_resource_snapshots_endpoint(
    api=api,
    resource_model=models.PluginTaskParameterType,
    resource_name=RESOURCE_TYPE,
    route_prefix=V1_PLUGIN_PARAMETER_TYPES_ROUTE,
    searchable_fields=SEARCHABLE_FIELDS,
    page_schema=PluginParameterTypePageSchema,
    build_fn=utils.build_plugin_parameter_type,
)
PluginParameterTypeSnapshotsIdResource = generate_resource_snapshots_id_endpoint(
    api=api,
    resource_model=models.PluginTaskParameterType,
    resource_name=RESOURCE_TYPE,
    response_schema=PluginParameterTypeSchema,
    build_fn=utils.build_plugin_parameter_type,
)

PluginParameterTypeTagsResource = generate_resource_tags_endpoint(
    api=api,
    resource_name=RESOURCE_TYPE,
)
PluginParameterTypeTagsIdResource = generate_resource_tags_id_endpoint(
    api=api,
    resource_name=RESOURCE_TYPE,
)



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from __future__ import annotations

from typing import cast

import numpy as np
import numpy.typing as npt
import structlog
from structlog.stdlib import BoundLogger

from .bounding_box_coordinates import BoundingBoxCoordinates, BoundingBoxesBatchedGrid

LOGGER: BoundLogger = structlog.stdlib.get_logger()

try:
    import tensorflow as tf
    from tensorflow import Tensor

except ImportError:  # pragma: nocover
    LOGGER.warn(
        "Unable to import one or more optional packages, functionality may be reduced",
        package="tensorflow",
    )


class TensorflowBoundingBoxCoordinates(BoundingBoxCoordinates):
    def __init__(self, grid_shape: tuple[int, int]) -> None:
        self._grid_shape = grid_shape
        self._cell_height = 1 / self._grid_shape[0]
        self._cell_width = 1 / self._grid_shape[1]
        self._cell_ncol = self._grid_shape[1]
        self._cell_nrow = self._grid_shape[0]

    @property
    def cell_height(self) -> float:
        return self._cell_height

    @property
    def cell_width(self) -> float:
        return self._cell_width

    @property
    def cell_ncol(self) -> int:
        return self._cell_ncol

    @property
    def cell_nrow(self) -> int:
        return self._cell_nrow

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.int32),
            tf.TensorSpec(None, tf.int32),
        ]
    )
    def apply_constraint_one_object_per_cell(
        self,
        bboxes_cell_xywh: Tensor,
        bboxes_cell_ij: Tensor,
        bboxes_labels: Tensor,
    ) -> tuple[Tensor, Tensor, Tensor]:
        return cast(
            tuple[Tensor, Tensor, Tensor],
            tf.numpy_function(
                self._apply_constraint_one_object_per_cell,
                [bboxes_cell_xywh, bboxes_cell_ij, bboxes_labels],
                [tf.float32, tf.int32, tf.int32],
            ),
        )

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.int32),
        ]
    )
    def find_no_obj_cell_ij(self, bboxes_cell_ij: Tensor) -> Tensor:
        i_range = tf.range(self.cell_nrow, dtype=tf.int32)
        j_range = tf.range(self.cell_ncol, dtype=tf.int32)

        return tf.numpy_function(
            self._find_no_obj_cell_ij, [bboxes_cell_ij, i_range, j_range], [tf.int32]
        )

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
        ]
    )
    def find_bbox_cell_ij(self, x_center: Tensor, y_center: Tensor) -> Tensor:
        cell_h = tf.cast(self.cell_height, tf.float32)
        cell_w = tf.cast(self.cell_width, tf.float32)
        max_i = tf.cast(
            tf.fill(dims=tf.shape(y_center), value=self.cell_nrow), tf.int32
        )
        max_j = tf.cast(
            tf.fill(dims=tf.shape(x_center), value=self.cell_ncol), tf.int32
        )

        i = tf.reduce_min(
            tf.stack(
                [
                    tf.cast(y_center / cell_h, tf.int32),
                    max_i,
                ],
                axis=-1,
            ),
            axis=-1,
        )
        j = tf.reduce_min(
            tf.stack(
                [
                    tf.cast(x_center / cell_w, tf.int32),
                    max_j,
                ],
                axis=-1,
            ),
            axis=-1,
        )

        return tf.cast(tf.stack([i, j], axis=-1), tf.int32)

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.int32),
        ]
    )
    def find_bbox_cell_xy(
        self, x_center: Tensor, y_center: Tensor, bboxes_cell_ij: Tensor
    ) -> Tensor:
        cell_h = tf.cast(self.cell_height, tf.float32)
        cell_w = tf.cast(self.cell_width, tf.float32)

        x_center_shape = tf.shape(x_center)
        y_center_shape = tf.shape(y_center)

        cell_x = tf.reduce_min(
            tf.stack(
                [
                    tf.cast(x_center / cell_w, tf.float32)
                    - tf.cast(bboxes_cell_ij[..., 1], tf.float32),
                    tf.ones(shape=x_center_shape, dtype=tf.float32),
                ],
                axis=-1,
            ),
            axis=-1,
        )
        cell_y = tf.reduce_min(
            tf.stack(
                [
                    tf.cast(y_center / cell_h, tf.float32)
                    - tf.cast(bboxes_cell_ij[..., 0], tf.float32),
                    tf.ones(shape=y_center_shape, dtype=tf.float32),
                ],
                axis=-1,
            ),
            axis=-1,
        )

        return cell_x, cell_y

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
        ]
    )
    def from_corner_to_image_xywh(self, bboxes_corner: Tensor) -> Tensor:
        xmin = bboxes_corner[..., 0]
        ymin = bboxes_corner[..., 1]
        xmax = bboxes_corner[..., 2]
        ymax = bboxes_corner[..., 3]

        x = (xmin + xmax) / 2.0
        y = (ymin + ymax) / 2.0
        w = xmax - xmin
        h = ymax - ymin

        return tf.stack([x, y, w, h], axis=-1)

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
        ]
    )
    def from_image_to_cell_xywh(
        self, bboxes_image_xywh: Tensor
    ) -> tuple[Tensor, Tensor]:
        x_center = bboxes_image_xywh[..., 0]
        y_center = bboxes_image_xywh[..., 1]

        bboxes_cell_ij = self.find_bbox_cell_ij(x_center=x_center, y_center=y_center)
        cell_x, cell_y = self.find_bbox_cell_xy(
            x_center=x_center, y_center=y_center, bboxes_cell_ij=bboxes_cell_ij
        )
        bboxes_cell_xywh = tf.stack(
            [cell_x, cell_y, bboxes_image_xywh[..., 2], bboxes_image_xywh[..., 3]],
            axis=-1,
        )

        return tf.cast(bboxes_cell_xywh, tf.float32), tf.cast(bboxes_cell_ij, tf.int32)

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
        ]
    )
    def from_corner_to_cell_xywh(self, bboxes_corner: Tensor) -> tuple[Tensor, Tensor]:
        bboxes_image_xywh = self.from_corner_to_image_xywh(bboxes_corner=bboxes_corner)
        bboxes_cell_xywh, bboxes_cell_ij = self.from_image_to_cell_xywh(
            bboxes_image_xywh=bboxes_image_xywh
        )

        return bboxes_cell_xywh, bboxes_cell_ij

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.int32),
        ]
    )
    def from_cell_xywh_to_image_xywh(
        self, bboxes_cell_xywh: Tensor, bboxes_cell_ij: Tensor
    ) -> Tensor:
        num_cells_wh = tf.concat(
            [tf.cast(self.cell_ncol, tf.float32), tf.cast(self.cell_nrow, tf.float32)],
            axis=0,
        )
        bboxes_image_xy = (
            bboxes_cell_xywh[..., :2]
            + tf.cast(tf.gather(bboxes_cell_ij, indices=[1, 0], axis=-1), tf.float32)
        ) / num_cells_wh
        bboxes_image_wh = bboxes_cell_xywh[..., 2:4]

        return tf.concat([bboxes_image_xy, bboxes_image_wh], axis=-1)

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
        ]
    )
    def from_image_xywh_to_corner(self, bboxes_image_xywh: Tensor) -> Tensor:
        x = bboxes_image_xywh[..., 0]
        y = bboxes_image_xywh[..., 1]
        w = bboxes_image_xywh[..., 2]
        h = bboxes_image_xywh[..., 3]

        x1 = x - (w / 2)
        x2 = x + (w / 2)
        y1 = y - (h / 2)
        y2 = y + (h / 2)

        return tf.stack([x1, y1, x2, y2], axis=-1)

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.int32),
        ]
    )
    def from_cell_xywh_to_corner(
        self, bboxes_cell_xywh: Tensor, bboxes_cell_ij: Tensor
    ) -> Tensor:
        bboxes_image_xywh = self.from_cell_xywh_to_image_xywh(
            bboxes_cell_xywh=bboxes_cell_xywh, bboxes_cell_ij=bboxes_cell_ij
        )
        bboxes_corner = self.from_image_xywh_to_corner(
            bboxes_image_xywh=bboxes_image_xywh
        )

        return bboxes_corner

    def _apply_constraint_one_object_per_cell(
        self,
        bboxes_cell_xywh: npt.NDArray,
        bboxes_cell_ij: npt.NDArray,
        bboxes_labels: npt.NDArray,
    ) -> tuple[npt.NDArray, npt.NDArray, npt.NDArray]:
        pruned_bboxes_cell_ij, pruning_indices = np.unique(
            bboxes_cell_ij, return_index=True, axis=0
        )
        pruned_bboxes_cell_xywh = bboxes_cell_xywh[pruning_indices]
        pruned_bboxes_labels = bboxes_labels[pruning_indices]

        return (
            pruned_bboxes_cell_xywh.astype("float32"),
            pruned_bboxes_cell_ij.astype("int32"),
            pruned_bboxes_labels.astype("int32"),
        )

    def _find_no_obj_cell_ij(
        self, bboxes_cell_ij: npt.NDArray, i_range: npt.NDArray, j_range: npt.NDArray
    ) -> npt.NDArray:
        cell_ij_set: set[tuple[npt.ArrayLike, npt.ArrayLike]] = set(
            [(i, j) for i in i_range for j in j_range]
        )
        cell_ij_seen: set[tuple[npt.ArrayLike, npt.ArrayLike]] = set(
            [(x[0], x[1]) for x in bboxes_cell_ij.tolist()]
        )
        no_obj_cell_ij: list[tuple[npt.ArrayLike, npt.ArrayLike]] = sorted(
            list(cell_ij_set - cell_ij_seen)
        )

        return np.array(no_obj_cell_ij, dtype="int32")


class TensorflowBoundingBoxesBatchedGrid(BoundingBoxesBatchedGrid):
    def __init__(
        self,
        bounding_box_coordinates: TensorflowBoundingBoxCoordinates,
    ) -> None:
        self._bbox_coord = bounding_box_coordinates
        self._wh_grid_indices = self._generate_wh_grid_indices(
            grid_shape=(
                bounding_box_coordinates.cell_nrow,
                bounding_box_coordinates.cell_ncol,
            )
        )
        self._num_wh_grid_indices = self._generate_num_wh_grid_indices(
            grid_shape=(
                bounding_box_coordinates.cell_nrow,
                bounding_box_coordinates.cell_ncol,
            )
        )

    @classmethod
    def on_grid_shape(
        cls, grid_shape: tuple[int, int]
    ) -> TensorflowBoundingBoxesBatchedGrid:
        return cls(
            bounding_box_coordinates=TensorflowBoundingBoxCoordinates(
                grid_shape=grid_shape
            ),
        )

    @property
    def cell_height(self) -> float:
        return self._bbox_coord.cell_height

    @property
    def cell_width(self) -> float:
        return self._bbox_coord.cell_width

    @property
    def cell_ncol(self) -> int:
        return self._bbox_coord.cell_ncol

    @property
    def cell_nrow(self) -> int:
        return self._bbox_coord.cell_nrow

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.int32),
            tf.TensorSpec(None, tf.int32),
        ]
    )
    def embed(
        self, bboxes_corner: Tensor, bboxes_labels: Tensor, n_classes: Tensor
    ) -> tuple[Tensor, Tensor, Tensor, Tensor]:
        (
            bboxes_cell_xywh_grid,
            bboxes_labels_grid,
            bboxes_object_mask,
            bboxes_no_object_mask,
        ) = tf.py_function(
            self._gen_embed_tensors,
            [n_classes],
            [tf.float32, tf.float32, tf.float32, tf.float32],
        )
        bboxes_cell_xywh, bboxes_cell_ij = self._bbox_coord.from_corner_to_cell_xywh(
            bboxes_corner=bboxes_corner
        )
        xywh, ij, labels = self._bbox_coord.apply_constraint_one_object_per_cell(
            bboxes_cell_xywh=bboxes_cell_xywh,
            bboxes_cell_ij=bboxes_cell_ij,
            bboxes_labels=bboxes_labels,
        )

        return cast(
            tuple[Tensor, Tensor, Tensor, Tensor],
            tf.numpy_function(
                self._embed,
                [
                    xywh,
                    ij,
                    labels,
                    bboxes_cell_xywh_grid,
                    bboxes_labels_grid,
                    bboxes_object_mask,
                    bboxes_no_object_mask,
                ],
                [tf.float32, tf.float32, tf.float32, tf.float32],
            ),
        )

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
        ]
    )
    def extract_using_mask(
        self,
        bboxes_grid: Tensor,
        labels_grid: Tensor,
        cell_mask: Tensor,
    ) -> tuple[Tensor, Tensor, Tensor]:
        bboxes_cell_ij = tf.where(cell_mask)
        bboxes = tf.gather_nd(params=bboxes_grid, indices=bboxes_cell_ij)
        labels = tf.gather_nd(params=labels_grid, indices=bboxes_cell_ij)

        return bboxes, bboxes_cell_ij, labels

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
        ]
    )
    def from_corner_to_image_xywh(self, bboxes_corner: Tensor) -> Tensor:
        return self._bbox_coord.from_corner_to_image_xywh(bboxes_corner=bboxes_corner)

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
        ]
    )
    def from_image_to_cell_xywh(self, bboxes_image_xywh: Tensor) -> Tensor:
        bboxes_cell_xywh, _ = self._bbox_coord.from_image_to_cell_xywh(
            bboxes_image_xywh=bboxes_image_xywh
        )

        return bboxes_cell_xywh

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
        ]
    )
    def from_corner_to_cell_xywh(self, bboxes_corner: Tensor) -> Tensor:
        bboxes_image_xywh: npt.NDArray = self.from_corner_to_image_xywh(
            bboxes_corner=bboxes_corner
        )
        bboxes_cell_xywh: npt.NDArray = self.from_image_to_cell_xywh(
            bboxes_image_xywh=bboxes_image_xywh
        )

        return bboxes_cell_xywh

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.int32),
        ]
    )
    def from_cell_xywh_to_image_xywh(
        self, bboxes_cell_xywh: Tensor, n_bounding_boxes: Tensor
    ) -> Tensor:
        batch_size = tf.shape(bboxes_cell_xywh)[0]

        wh_grid = tf.tile(
            tf.cast(self._wh_grid_indices, tf.float32),
            multiples=[batch_size, 1, 1, n_bounding_boxes, 1],
        )
        num_wh_grid = tf.tile(
            tf.cast(self._num_wh_grid_indices, tf.float32),
            multiples=[batch_size, 1, 1, n_bounding_boxes, 1],
        )

        bboxes_image_xy = (bboxes_cell_xywh[..., :2] + wh_grid) / num_wh_grid
        bboxes_image_wh = bboxes_cell_xywh[..., 2:4]

        return tf.concat([bboxes_image_xy, bboxes_image_wh], axis=-1)

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
        ]
    )
    def from_image_xywh_to_corner(self, bboxes_image_xywh: Tensor) -> Tensor:
        return self._bbox_coord.from_image_xywh_to_corner(
            bboxes_image_xywh=bboxes_image_xywh
        )

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.int32),
        ]
    )
    def from_cell_xywh_to_corner(
        self, bboxes_cell_xywh: Tensor, n_bounding_boxes: Tensor
    ) -> Tensor:
        bboxes_image_xywh = self.from_cell_xywh_to_image_xywh(
            bboxes_cell_xywh=bboxes_cell_xywh,
            n_bounding_boxes=n_bounding_boxes,
        )
        bboxes_corner = self.from_image_xywh_to_corner(
            bboxes_image_xywh=bboxes_image_xywh
        )

        return bboxes_corner

    def _embed(
        self,
        xywh: npt.NDArray,
        ij: npt.NDArray,
        labels: npt.NDArray,
        bboxes_cell_xywh_grid: npt.NDArray,
        bboxes_labels_grid: npt.NDArray,
        bboxes_object_mask: npt.NDArray,
        bboxes_no_object_mask: npt.NDArray,
    ) -> tuple[npt.NDArray, npt.NDArray, npt.NDArray, npt.NDArray]:
        grid_i = np.transpose(ij)[0]
        grid_j = np.transpose(ij)[1]

        bboxes_cell_xywh_grid[grid_i, grid_j] = xywh
        bboxes_labels_grid[grid_i, grid_j, labels] = 1.0
        bboxes_object_mask[grid_i, grid_j] = 1.0
        bboxes_no_object_mask[grid_i, grid_j] = 0.0

        bboxes_cell_xywh_grid = np.expand_dims(bboxes_cell_xywh_grid, axis=-2)

        return (
            bboxes_cell_xywh_grid,
            bboxes_labels_grid,
            bboxes_object_mask,
            bboxes_no_object_mask,
        )

    def _gen_embed_tensors(
        self, n_classes: Tensor
    ) -> tuple[Tensor, Tensor, Tensor, Tensor]:
        bboxes_cell_xywh_grid = tf.zeros(
            shape=(self.cell_nrow, self.cell_ncol, 4),
            dtype=tf.float32,
        )
        bboxes_labels_grid = tf.zeros(
            shape=(self.cell_nrow, self.cell_ncol, n_classes),
            dtype=tf.float32,
        )
        bboxes_object_mask = tf.zeros(
            shape=(self.cell_nrow, self.cell_ncol),
            dtype=tf.float32,
        )
        bboxes_no_object_mask = tf.ones(
            shape=(self.cell_nrow, self.cell_ncol),
            dtype=tf.float32,
        )

        return (
            bboxes_cell_xywh_grid,
            bboxes_labels_grid,
            bboxes_object_mask,
            bboxes_no_object_mask,
        )

    @staticmethod
    def _generate_wh_grid_indices(grid_shape: tuple[int, int]) -> Tensor:
        w_grid_indices = tf.tile(
            tf.expand_dims(tf.range(grid_shape[1], dtype=tf.int32), axis=0),
            multiples=[grid_shape[0], 1],
        )
        w_grid_indices = tf.reshape(
            w_grid_indices, (1, grid_shape[0], grid_shape[1], 1, 1)
        )

        h_grid_indices = tf.transpose(
            tf.tile(
                tf.expand_dims(tf.range(grid_shape[0], dtype=tf.int32), axis=0),
                multiples=[grid_shape[1], 1],
            )
        )
        h_grid_indices = tf.reshape(
            h_grid_indices, (1, grid_shape[0], grid_shape[1], 1, 1)
        )

        return tf.cast(tf.concat([w_grid_indices, h_grid_indices], axis=-1), tf.float32)

    @staticmethod
    def _generate_num_wh_grid_indices(grid_shape: tuple[int, int]) -> Tensor:
        num_w_grid_indices = tf.constant(
            value=grid_shape[1],
            shape=(1, grid_shape[0], grid_shape[1], 1, 1),
            dtype=tf.int32,
        )
        num_h_grid_indices = tf.constant(
            value=grid_shape[0],
            shape=(1, grid_shape[0], grid_shape[1], 1, 1),
            dtype=tf.int32,
        )

        return tf.cast(
            tf.concat([num_w_grid_indices, num_h_grid_indices], axis=-1), tf.float32
        )



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from pathlib import Path
from typing import Union


def set_path_ext(filepath: Union[str, Path], ext: str) -> Path:
    ext = f".{ext.lstrip('.')}"
    filename: Union[str, Path] = Path(filepath)
    filepath = Path(filepath)

    for _ in filepath.suffixes:
        filename = Path(filename).stem

    return filepath.with_name(f"{Path(filename).with_suffix(ext)}")
